import asyncio
import logging

import pytest

from aioafero import EventType, InvalidAuth, AferoDevice, AferoState, TemperatureUnit
from aioafero.errors import DeviceNotFound, AferoError, ExceededMaximumRetries
from aioafero.v1 import AferoBridgeV1, TokenData, add_secret, v1_const, TemperatureUnit
from aioafero.v1.controllers.device import DeviceController
from aioafero.v1.controllers.event import EventStream
from aioafero.v1.controllers.fan import FanController
from aioafero.v1.controllers.light import LightController
from aioafero.v1.controllers.lock import LockController
from aioafero.v1.controllers.switch import SwitchController
from aioafero.v1.controllers.valve import ValveController
from aiohttp.client_exceptions import ClientResponseError
from dataclasses import asdict
from aioafero.v1 import auth
import json
from urllib.parse import urlencode

from . import utils


async def build_url(base_url: str, qs: dict[str, str]) -> str:
    return f"{base_url}?{urlencode(qs)}"


zandra_light = utils.create_devices_from_data("fan-ZandraFan.json")[1]


@pytest.mark.skip(reason="Not yet implemented")
def test_context_manager(mocked_bridge):
    pass


def test_devices(mocked_bridge):
    assert isinstance(mocked_bridge.devices, DeviceController)


def test_events(mocked_bridge):
    assert isinstance(mocked_bridge.events, EventStream)


def test_fans(mocked_bridge):
    assert isinstance(mocked_bridge.fans, FanController)


def test_lights(mocked_bridge):
    assert isinstance(mocked_bridge.lights, LightController)


def test_locks(mocked_bridge):
    assert isinstance(mocked_bridge.locks, LockController)


def test_switches(mocked_bridge):
    assert isinstance(mocked_bridge.switches, SwitchController)


def test_valves(mocked_bridge):
    assert isinstance(mocked_bridge.valves, ValveController)


def test_controllers(mocked_bridge):
    mocked_bridge.devices._initialized = False
    mocked_bridge.exhaust_fans._initialized = False
    mocked_bridge.fans._initialized = False
    mocked_bridge.lights._initialized = False
    mocked_bridge.locks._initialized = True
    mocked_bridge.portable_acs._initialized = False
    mocked_bridge.security_systems._initialized = False
    mocked_bridge.security_systems_keypads._initialized = False
    mocked_bridge.security_systems_sensors._initialized = False
    mocked_bridge.switches._initialized = False
    mocked_bridge.thermostats._initialized = False
    mocked_bridge.valves._initialized = True
    assert mocked_bridge.controllers == [mocked_bridge.locks, mocked_bridge.valves]
    mocked_bridge.switches._initialized = True
    assert mocked_bridge.controllers == [
        mocked_bridge.locks,
        mocked_bridge.switches,
        mocked_bridge.valves,
    ]


def test_tracked_devices(mocked_bridge):
    assert mocked_bridge.tracked_devices == set()
    mocked_bridge.add_device(zandra_light.id, mocked_bridge.lights)
    assert mocked_bridge.tracked_devices == {zandra_light.id}


def test_add_device(mocked_bridge):
    assert mocked_bridge.tracked_devices == set()
    mocked_bridge.add_device(zandra_light.id, mocked_bridge.lights)
    assert mocked_bridge.tracked_devices == {zandra_light.id}
    assert mocked_bridge._known_devs == {zandra_light.id: mocked_bridge.lights}


def test_remove_device(mocked_bridge):
    assert mocked_bridge.tracked_devices == set()
    mocked_bridge.add_device(zandra_light.id, mocked_bridge.lights)
    assert mocked_bridge.tracked_devices == {zandra_light.id}
    assert mocked_bridge._known_devs == {zandra_light.id: mocked_bridge.lights}
    mocked_bridge.remove_device(zandra_light.id)
    assert mocked_bridge.tracked_devices == set()
    assert mocked_bridge._known_devs == {}


def test_set_polling_interval(mocked_bridge):
    assert mocked_bridge.events._polling_interval == 30
    mocked_bridge.set_polling_interval(10)
    assert mocked_bridge.events._polling_interval == 10


@pytest.mark.asyncio
async def test_initialize(bridge_with_acct, mocker):
    mocker.patch.object(bridge_with_acct, "request")
    mocker.patch.object(bridge_with_acct, "get_account_id", return_value="mocked-account-id")
    mocker.patch.object(bridge_with_acct.events, "wait_for_first_poll")
    mocker.patch.object(bridge_with_acct.devices, "_initialized", True)
    await bridge_with_acct.initialize()
    await bridge_with_acct.async_block_until_done()

@pytest.mark.asyncio
@pytest.mark.parametrize(
    "expected_val,error,temperature_unit",
    [
        # good data
        ([], False, TemperatureUnit.FAHRENHEIT),
        ([], False, TemperatureUnit.CELSIUS),
        # bad data
        ("i dont know", True, TemperatureUnit.CELSIUS),
    ],
)
async def test_fetch_data(expected_val, error, temperature_unit, mocked_bridge_req, mocker):
    expected = mocker.Mock()
    mocked_bridge_req.temperature_unit = temperature_unit
    mocker.patch.object(
        expected, "json", side_effect=mocker.AsyncMock(return_value=expected_val)
    )
    mocker.patch.object(mocked_bridge_req, "request", return_value=expected)
    if not error:
        assert await mocked_bridge_req.fetch_data() == expected_val
        call = mocked_bridge_req.request.call_args
        params = call[1]["params"]
        if temperature_unit == TemperatureUnit.FAHRENHEIT:
            assert "units" in params
            assert params["units"] == TemperatureUnit.FAHRENHEIT.value
        else:
            assert "units" not in params
    else:
        with pytest.raises(TypeError):
            await mocked_bridge_req.fetch_data()


def fake_version_data(*args, **kwargs):
    yield {"version": "1.0.0"}
    yield {"version": "2.0.0"}


@pytest.mark.asyncio
async def test_fetch_data_with_version(mocked_bridge_req, mocker):
    get_device_versions = mocker.patch.object(mocked_bridge_req, "get_device_version", side_effect=fake_version_data())
    mocked_response = [
        {
            "typeId": "metadevice.room"
        },
        {
            "typeId": "metadevice.device",
            "deviceId": "test_device_id",
        },
        {
            "typeId": "metadevice.device",
            "deviceId": "test_device_id2",
        },
        {
            "typeId": "metadevice.device",
            "deviceId": "test_device_id",
        },
    ]
    expected = mocker.Mock()
    mocker.patch.object(
        expected, "json", side_effect=mocker.AsyncMock(return_value=mocked_response)
    )
    mocker.patch.object(mocked_bridge_req, "request", return_value=expected)
    resp = await mocked_bridge_req.fetch_data(version_poll=True)
    assert get_device_versions.call_count == 2
    assert get_device_versions.call_args_list[0][0][0] == "test_device_id"
    assert get_device_versions.call_args_list[1][0][0] == "test_device_id2"
    assert resp == [
        {'typeId': 'metadevice.room'},
        {'typeId': 'metadevice.device', 'deviceId': 'test_device_id', 'version_data': {'version': '1.0.0'}},
        {'typeId': 'metadevice.device', 'deviceId': 'test_device_id2', 'version_data': {'version': '2.0.0'}},
        {'typeId': 'metadevice.device', 'deviceId': 'test_device_id', 'version_data': {'version': '1.0.0'}}
    ]


@pytest.mark.asyncio
async def test_get_device_versions(mocked_bridge, mocker):
    req = mocker.patch.object(mocked_bridge, "request")
    await mocked_bridge.get_device_version("test_device_id")
    req.assert_called_once_with("GET", "https://api2.afero.net/v1/accounts/mocked-account-id/devices/test_device_id/versions")


@pytest.mark.asyncio
async def test_send_service_request_dev_not_found(mocked_bridge):
    with pytest.raises(DeviceNotFound):
        await mocked_bridge.send_service_request("no", [{}])


@pytest.mark.asyncio
async def test_send_service_request(mocked_bridge, mocker):
    controller = mocked_bridge.lights
    await mocked_bridge.events.generate_events_from_data(
        [utils.create_hs_raw_from_device(zandra_light)]
    )
    await mocked_bridge.async_block_until_done()
    mocked_bridge.add_device(zandra_light.id, controller)
    assert controller[zandra_light.id].on.on is True
    states = [{"functionClass": "power", "functionInstance": "light-power", "value": "off"}]
    resp = mocker.AsyncMock()
    resp.json = mocker.AsyncMock(return_value={"metadeviceId": zandra_light.id, "values": states})
    mocker.patch.object(
        controller, "update_afero_api", return_value=resp
    )
    await mocked_bridge.send_service_request(
        zandra_light.id,
        states,
    )
    await mocked_bridge.async_block_until_done()
    assert controller[zandra_light.id].on.on is False


@pytest.mark.asyncio
async def test_create_request_err(mocked_bridge, mocker):
    mocker.patch.object(mocked_bridge._auth, "token", side_effect=InvalidAuth)
    emit = mocker.patch.object(mocked_bridge.events, "emit")
    with pytest.raises(InvalidAuth):
        async with mocked_bridge.create_request("get", "https://not-called.io", True):
            pass

    emit.assert_called_once_with(EventType.INVALID_AUTH)


@pytest.mark.parametrize(
    "hide_secrets",
    [True, False],
)
def test_AferoBridgeV1_hide_secrets(hide_secrets, caplog):
    caplog.set_level(logging.DEBUG)
    bridge = AferoBridgeV1("username", "password", hide_secrets=hide_secrets)
    secret = "this-is-super-secret-beans"
    add_secret(secret)
    with bridge.secret_logger():
        bridge.logger.debug(secret)
    if hide_secrets:
        assert "th***ns" in caplog.text
    else:
        assert secret in caplog.text


@pytest.mark.asyncio
@pytest.mark.parametrize(
    ("account_id", "response", "expected", "err"),
    [
        # Uninitialized, happy patch
        (
            None,
            {
                "status": 200,
                "payload": {
                    "accountAccess": [
                        {
                            "account": {
                                "accountId": "test_account_id"
                            }
                        }
                    ]
                }
            },
            "test_account_id",
            None,
        ),
        # Uninitialized, bad status
        (
            None,
            {
                "status": 400,
                "payload": {
                    "accountAccess": [
                        {
                            "account": {
                                "accountId": "test_account_id"
                            }
                        }
                    ]
                }
            },
            None,
            ClientResponseError,
        ),
        # Uninitialized, bad response
        (
            None,
            {
                "status": 200,
                "payload": {
                }
            },
            None,
            AferoError,
        ),
        # Initialized, dont do anything
        (
            "mocked-account-id",
            None,
            "mocked-account-id",
            None,
        )
    ]
)
async def test_get_account_id(account_id, response, expected, err, mock_aioresponse, bridge_with_acct, mocker):
    bridge = bridge_with_acct
    mocker.patch.object(bridge, "_account_id", account_id)
    if account_id is None:
        url = bridge.generate_api_url(v1_const.AFERO_GENERICS["ACCOUNT_ID_ENDPOINT"])
        mock_aioresponse.get(url, payload=response["payload"], status=response["status"])
        if not err:
            assert await bridge.get_account_id() == expected
        else:
            with pytest.raises(err):
                await bridge.get_account_id()
    else:
        logger = mocker.patch.object(bridge_with_acct, "logger")
        await bridge.get_account_id()
        logger.assert_not_called()


class DummyResponse:
    def __init__(self, **kwargs):
        for k, v in kwargs.items():
            setattr(self, k, v)

    async def read(self):
        return "cool-beans"


def test_set_token_data(mocked_bridge):
    data = TokenData(
        "token",
        "access_token",
        "refresh_token",
        12345,
    )
    mocked_bridge.set_token_data(data)
    assert mocked_bridge.refresh_token == "refresh_token"


@pytest.mark.asyncio
async def test_cleanup_process(mocked_bridge_req):
    mocked_bridge_req.add_job(asyncio.create_task(asyncio.sleep(1)))
    assert len(mocked_bridge_req._adhoc_tasks) == 1
    await mocked_bridge_req.initialize_cleanup()
    await mocked_bridge_req.async_block_until_done()
    assert len(mocked_bridge_req._adhoc_tasks) == 0
    await mocked_bridge_req.close()


@pytest.mark.asyncio
async def test_double_initialize(bridge):
    assert len(bridge._scheduled_tasks) == 1
    await bridge.initialize()
    assert len(bridge._scheduled_tasks) == 1


class AsyncContextManagerMock:
    def __init__(self):
        self.total = 0

    async def __aenter__(self):
        status_code = 429 if self.total < 2 else 200
        self.total += 1
        return DummyResponse(status=status_code)

    async def __aexit__(self, exc_type, exc, tb):
        pass



@pytest.mark.asyncio
@pytest.mark.parametrize(
    "max_retries, times_to_sleep, response_gen, exp_error", [
        # Out of retries
        (
            0, 0, None, ExceededMaximumRetries
        ),
        # Double retry
        (
            None, 2, True, None
        )
    ]
)
async def test_request(max_retries, times_to_sleep, response_gen, exp_error, mocker):
    bridge = AferoBridgeV1("username", "password")
    mock_sleep = mocker.patch("asyncio.sleep", new=mocker.AsyncMock())
    if response_gen:
        async_mock_response = AsyncContextManagerMock()
        mocker.patch.object(bridge, "create_request", return_value=async_mock_response)
    if max_retries is not None:
        mocker.patch.object(v1_const, "MAX_RETRIES", max_retries)
    if exp_error:
        with pytest.raises(exp_error):
            await bridge.request("fff", "fff")
    else:
        await bridge.request("fff", "fff")
    if times_to_sleep:
        assert mock_sleep.call_count == times_to_sleep


def test_get_afero_device(mocked_bridge):
    with pytest.raises(DeviceNotFound):
        mocked_bridge.get_afero_device("nope")


def test_fetch_device_states(mocked_bridge, mocker):
    states = [
        AferoState(functionClass="power", functionInstance="light-power", value="on"),
        AferoState(functionClass="brightness", functionInstance="light-brightness", value="100"),
    ]
    dummy_dev = AferoDevice(
        id="beans",
        device_id="parent-bean",
        device_class="bean-class",
        default_name="beans",
        default_image="beans",
        friendly_name="Beans",
        model="bean-model",
        manufacturerName="bean-co",
        states=states,
        functions=[],
    )
    resp_states = [asdict(x) for x in states]
    resp_states.append({"functionClass": "unknown", "functionInstance": "unknown"})
    json_resp = mocker.AsyncMock()
    json_resp.return_value = {"metadeviceId": dummy_dev.id, "values": resp_states}
    resp = mocker.AsyncMock()
    resp.json = json_resp
    resp.status = 200
    mocker.patch.object(mocked_bridge, "request", return_value=resp)
    states = asyncio.run(mocked_bridge.fetch_device_states(dummy_dev.id))
    assert states == dummy_dev.states


def test_get_device_controller(mocked_bridge):
    mocked_bridge.add_device(zandra_light.id, mocked_bridge.lights)
    assert mocked_bridge.get_device_controller(zandra_light.id) == mocked_bridge.lights
    with pytest.raises(DeviceNotFound):
        mocked_bridge.get_device_controller("nope")


@pytest.mark.asyncio
async def test_otp_login(mock_aioresponse, aio_sess, mocker, mocked_bridge_req, caplog):
    caplog.set_level(0)
    mocked_bridge_req._web_session = aio_sess
    hs_auth = mocked_bridge_req._auth
    challenge = await hs_auth.generate_challenge_data()
    hs_auth._bridge._web_session = aio_sess
    auth_sess_data = auth.AuthSessionData("url_sess_code", "url_exec_code", "url_tab_id")
    url_params = auth.extract_login_codes(auth_sess_data, hs_auth._afero_client)
    hs_auth._otp_data = {
        "params": url_params,
        "headers": {},
        "challenge": challenge,
    }
    url = hs_auth.generate_auth_url(v1_const.AFERO_GENERICS["AUTH_CODE_ENDPOINT"])
    url = await build_url(url, url_params)
    # Successful OTP POST
    otp_post_response = {
        "status": 302,
        "headers": {
            "location": (
                "hubspace-app://loginredirect"
                "?session_state=sess-state"
                "&iss=https%3A%2F%2Faccounts.hubspaceconnect.com"
                "%2Fauth%2Frealms%2Fthd&code=code"
            )
        },
    }
    mock_aioresponse.post(url, **otp_post_response)
    # Successful authorization_code generation
    url = hs_auth.generate_auth_url(v1_const.AFERO_GENERICS["AUTH_TOKEN_ENDPOINT"])
    resp_data = {
        "refresh_token": "refresh_token",
        "access_token": "access_token",
        "id_token": "id_token",
    }
    mock_aioresponse.post(url, status=200, body=json.dumps(resp_data))
    await mocked_bridge_req.otp_login("123456")
    assert mocked_bridge_req._auth._token_data == auth.TokenData(
        token="id_token",
        access_token="access_token",
        refresh_token="refresh_token",
        expiration=mocker.ANY,
    )


@pytest.mark.asyncio
def test_unsubscribe():
    bridge = AferoBridgeV1("username", "password")

    def whatever(*args, **kwargs):
        pass

    # Ensure at least two controllers are initialized to validate unsubscribe functionality
    bridge.devices._initialized = True
    bridge.fans._initialized = True
    unsub = bridge.subscribe(whatever)
    assert bridge.devices._subscribers == {"*": [(whatever, None)]}
    assert bridge.fans._subscribers == {"*": [(whatever, None)]}
    assert callable(unsub)
    unsub()
    assert bridge.devices._subscribers == {"*": []}
    assert bridge.fans._subscribers == {"*": []}


@pytest.mark.parametrize(
    ("start_unit", "new_unit", "called"), [
        # No change
        (TemperatureUnit.CELSIUS, TemperatureUnit.CELSIUS, False),
        # Change detected
        (TemperatureUnit.CELSIUS, TemperatureUnit.FAHRENHEIT, True),
    ]
)
async def test_adjust_temperature_unit(start_unit, new_unit, called, mocked_bridge, mocker):
    mocker.patch.object(mocked_bridge, "temperature_unit", start_unit)
    mocker.patch.object(mocked_bridge, "add_job", side_effect=mocked_bridge.add_job)
    await mocked_bridge.adjust_temperature_unit(new_unit)
    assert mocked_bridge.temperature_unit == new_unit
    assert mocked_bridge.add_job.called == called



"""Controller holding and managing Afero IoT resources of type `security-system`."""

import asyncio
import copy

from aioafero.device import (
    AferoCapability,
    AferoDevice,
    AferoState,
    get_function_from_device,
)
from aioafero.errors import DeviceNotFound, SecuritySystemError
from aioafero.util import process_function
from aioafero.v1.models import SecuritySystem, SecuritySystemPut, features
from aioafero.v1.models.resource import DeviceInformation, ResourceTypes

from .base import AferoBinarySensor, AferoSensor, BaseResourcesController, NumbersName
from .event import CallbackResponse

SENSOR_SPLIT_IDENTIFIER = "sensor"
GENERIC_MODES = {0: "Off", 1: "On"}
TRIGGER_MODES = {
    0: "Off",
    1: "Home",
    2: "Away",
    3: "Home/Away",
}
KNOWN_SENSOR_MODELS = {
    1: "Motion Sensor",
    2: "Door/Window Sensor",
}
BYPASS_MODES = {
    0: "Off",
    1: "Manual",
    4: "On",
}

UPDATE_TIME = 3  # Seconds after performing an action to refresh state


def get_sensor_ids(device) -> set[int]:
    """Determine available sensors from the states."""
    sensor_ids = set()
    for state in device.states:
        if state.functionInstance is None:
            continue
        if state.functionInstance.startswith("sensor-") and state.value is not None:
            sensor_id = int(state.functionInstance.split("-", 1)[1])
            sensor_ids.add(sensor_id)
    return sensor_ids


def generate_sensor_name(afero_device, sensor_id: int) -> str:
    """Generate the name for an instanced element."""
    return f"{afero_device.id}-{SENSOR_SPLIT_IDENTIFIER}-{sensor_id}"


def get_valid_states(afero_states: list, sensor_id: int) -> list:
    """Find states associated with the specific sensor."""
    valid_states: list = []
    for state in afero_states:
        if (
            state.functionClass not in ["sensor-state", "sensor-config"]
            or state.value is None
        ):
            continue
        state_sensor_split = state.functionInstance.rsplit("-", 1)
        state_sensor_id = int(state_sensor_split[1])
        if state_sensor_id != sensor_id:
            continue
        top_level_key = list(state.value.keys())[0]
        if state.functionClass == "sensor-state":
            valid_states.append(
                AferoState(
                    functionClass="battery-level",
                    functionInstance=None,
                    value=state.value[top_level_key]["batteryLevel"],
                )
            )
            valid_states.append(
                AferoState(
                    functionClass="tampered",
                    functionInstance=None,
                    value=GENERIC_MODES[state.value[top_level_key]["tampered"]],
                )
            )
            valid_states.append(
                AferoState(
                    functionClass="triggered",
                    functionInstance=None,
                    value=GENERIC_MODES[state.value[top_level_key]["triggered"]],
                )
            )
            valid_states.append(
                AferoState(
                    functionClass="available",
                    functionInstance=None,
                    value=not bool(state.value[top_level_key]["missing"]),
                )
            )
        else:
            valid_states.append(
                AferoState(
                    functionClass="chirpMode",
                    functionInstance=None,
                    value=GENERIC_MODES[state.value[top_level_key]["chirpMode"]],
                )
            )
            valid_states.append(
                AferoState(
                    functionClass="triggerType",
                    functionInstance=None,
                    value=TRIGGER_MODES[state.value[top_level_key]["triggerType"]],
                )
            )
            valid_states.append(
                AferoState(
                    functionClass="bypassType",
                    functionInstance=None,
                    value=BYPASS_MODES[state.value[top_level_key]["bypassType"]],
                )
            )
            valid_states.append(
                AferoState(
                    functionClass="top-level-key",
                    functionInstance=None,
                    value=top_level_key,
                )
            )
    return valid_states


def get_model_type(states: list[AferoState], sensor_id: int) -> str:
    """Get the model type from the state list."""
    for state in states:
        if state.functionClass not in ["sensor-state"] or state.value is None:
            continue
        state_sensor_split = state.functionInstance.rsplit("-", 1)
        state_sensor_id = int(state_sensor_split[1])
        if state_sensor_id != sensor_id:
            continue
        top_level_key = list(state.value.keys())[0]
        return KNOWN_SENSOR_MODELS.get(
            int(state.value[top_level_key]["deviceType"]), "Unknown"
        )
    return "Unknown"


def get_valid_functions(afero_functions: list, sensor_id: int) -> list:
    """Find functions associated with the specific sensor."""
    valid_functions: list = []
    for func in afero_functions:
        if func["functionClass"] not in ["sensor-state", "sensor-config"]:
            continue
        sensor_split = func["functionInstance"].rsplit("-", 1)
        state_sensor_id = int(sensor_split[1])
        if state_sensor_id != sensor_id:
            continue
        if func["functionClass"] == "sensor-config":
            valid_functions.append(
                {
                    "functionClass": "chirpMode",
                    "functionInstance": func["functionInstance"],
                    "type": "category",
                    "values": [{"name": x} for x in GENERIC_MODES.values()],
                }
            )
            valid_functions.append(
                {
                    "functionClass": "triggerType",
                    "functionInstance": func["functionInstance"],
                    "type": "category",
                    "values": [{"name": x} for x in TRIGGER_MODES.values()],
                }
            )
            valid_functions.append(
                {
                    "functionClass": "bypassType",
                    "functionInstance": func["functionInstance"],
                    "type": "category",
                    "values": [{"name": x} for x in BYPASS_MODES.values()],
                }
            )
    return valid_functions


def get_sensor_name(afero_capabilities: list[AferoCapability], sensor_id: int) -> str:
    """Get the Afero name for a specific sensor."""
    for capability in afero_capabilities:
        if (
            capability.functionClass != "sensor-state"
            or capability.functionInstance != f"sensor-{sensor_id}"
            or capability.options.get("name") is None
        ):
            continue
        return capability.options["name"]
    return f"Sensor {sensor_id}"


def security_system_callback(afero_device: AferoDevice) -> CallbackResponse:
    """Convert an AferoDevice into multiple devices."""
    multi_devs: list[AferoDevice] = []
    if afero_device.device_class == "security-system":
        for sensor_id in get_sensor_ids(afero_device):
            cloned = copy.deepcopy(afero_device)
            cloned.device_id = generate_sensor_name(afero_device, sensor_id)
            cloned.id = generate_sensor_name(afero_device, sensor_id)
            cloned.split_identifier = SENSOR_SPLIT_IDENTIFIER
            cloned.device_class = ResourceTypes.SECURITY_SYSTEM_SENSOR.value
            cloned.friendly_name = f"{afero_device.friendly_name} - {get_sensor_name(afero_device.capabilities, sensor_id)}"
            cloned.states = get_valid_states(afero_device.states, sensor_id)
            cloned.functions = get_valid_functions(afero_device.functions, sensor_id)
            cloned.model = f"{afero_device.model} - {get_model_type(afero_device.states, sensor_id)}"
            multi_devs.append(cloned)
    return CallbackResponse(
        split_devices=multi_devs,
        remove_original=False,
    )


class SecuritySystemController(BaseResourcesController[SecuritySystem]):
    """Controller holding and managing Afero IoT resources of type `security-system`."""

    ITEM_TYPE_ID = ResourceTypes.DEVICE
    ITEM_TYPES = [ResourceTypes.SECURITY_SYSTEM]
    ITEM_CLS = SecuritySystem
    ITEM_MAPPING = {"alarm_state": "alarm-state"}
    # Sensors map functionClass -> Unit
    ITEM_SENSORS: dict[str, str] = {
        "alarm-state": None,
        "history-event": None,
        "disarmed-by": None,
    }
    # Binary sensors map key -> alerting value
    ITEM_BINARY_SENSORS: dict[str, str] = {
        "battery-powered": "battery-powered",
    }
    # Elements that map to numbers. func class / func instance to NumbersName
    ITEM_NUMBERS: dict[tuple[str, str | None], NumbersName] = {
        ("arm-exit-delay", "away"): NumbersName(
            unit="seconds", display_name="Arm Exit Delay Away"
        ),
        ("arm-exit-delay", "stay"): NumbersName(
            unit="seconds", display_name="Arm Exit Delay Home"
        ),
        ("temporary-bypass-time", None): NumbersName(
            unit="seconds", display_name="Bypass Time"
        ),
        ("disarm-entry-delay", None): NumbersName(
            unit="seconds", display_name="Disarm Entry Delay"
        ),
        ("siren-alarm-timeout", None): NumbersName(
            unit="seconds", display_name="Siren Timeout"
        ),
    }
    # Elements that map to Select. func class / func instance to name
    ITEM_SELECTS = {
        ("song-id", "alarm"): "Alarm Noise",
        ("volume", "siren"): "Alarm Volume",
        ("bypass-allowed", None): "Enable Temporary Bypass",
        ("song-id", "chime"): "Chime Noise",
        ("volume", "chime"): "Chime Volume",
        ("volume", "entry-delay"): "Entry Delay Volume",
        ("volume", "exit-delay-away"): "Exit Delay Volume Away",
        ("volume", "exit-delay-stay"): "Exit Delay Volume Home",
        # ("dark-mode", None): "KeyPad Dark Mode",
        # ("song-id", "beep"): "KeyPad Noise",
    }
    # Split sensors from the primary payload
    DEVICE_SPLIT_CALLBACKS: dict[str, callable] = {
        ResourceTypes.SECURITY_SYSTEM_SENSOR.value: security_system_callback
    }

    async def disarm(self, device_id: str, disarm_pin: int) -> None:
        """Disarm the system."""
        await self.set_state(device_id, disarm_pin=disarm_pin)

    async def arm_home(self, device_id: str) -> None:
        """Arms the system while someone is home."""
        await self.set_state(device_id, command=4)

    async def arm_away(self, device_id: str) -> None:
        """Arms the system while no one is home."""
        await self.set_state(device_id, command=2)

    async def alarm_trigger(self, device_id: str) -> None:
        """Manually trigger the alarm."""
        await self.set_state(device_id, command=5)

    async def initialize_elem(self, afero_device: AferoDevice) -> SecuritySystem:
        """Initialize the element.

        :param afero_device: Afero Device that contains the updated states

        :return: Newly initialized resource
        """
        available: bool = False
        alarm_state: features.ModeFeature | None = None
        siren_action: features.SecuritySensorSirenFeature | None = None
        numbers: dict[tuple[str, str | None], features.NumbersFeature] | None = {}
        selects: dict[tuple[str, str | None], features.SelectFeature] | None = {}
        sensors: dict[str, AferoSensor] = {}
        binary_sensors: dict[str, AferoBinarySensor] = {}
        for state in afero_device.states:
            func_def = get_function_from_device(
                afero_device.functions, state.functionClass, state.functionInstance
            )
            if state.functionClass == "available":
                available = state.value
            elif state.functionClass == "alarm-state":
                alarm_state = features.ModeFeature(
                    mode=state.value,
                    modes=set(
                        process_function(afero_device.functions, state.functionClass)
                    ),
                )
            elif sensor := await self.initialize_sensor(state, afero_device.device_id):
                if isinstance(sensor, AferoBinarySensor):
                    binary_sensors[sensor.id] = sensor
                else:
                    sensors[sensor.id] = sensor
            elif number := await self.initialize_number(func_def, state):
                numbers[number[0]] = number[1]
            elif select := await self.initialize_select(afero_device.functions, state):
                selects[select[0]] = select[1]
            elif state.functionClass == "siren-action":
                try:
                    result_code = state.value["security-siren-action"]["resultCode"]
                    command = state.value["security-siren-action"]["command"]
                except TypeError:
                    result_code = None
                    command = None
                siren_action = features.SecuritySensorSirenFeature(
                    result_code=result_code,
                    command=command,
                )

        self._items[afero_device.id] = SecuritySystem(
            _id=afero_device.id,
            available=available,
            sensors=sensors,
            binary_sensors=binary_sensors,
            numbers=numbers,
            selects=selects,
            device_information=DeviceInformation(
                device_class=afero_device.device_class,
                default_image=afero_device.default_image,
                default_name=afero_device.default_name,
                manufacturer=afero_device.manufacturerName,
                model=afero_device.model,
                name=afero_device.friendly_name,
                parent_id=afero_device.device_id,
                children=afero_device.children,
                functions=afero_device.functions,
            ),
            alarm_state=alarm_state,
            siren_action=siren_action,
        )
        return self._items[afero_device.id]

    async def update_elem(self, afero_device: AferoDevice) -> set:
        """Update the Security System with the latest API data.

        :param afero_device: Afero Device that contains the updated states

        :return: States that have been modified
        """
        cur_item = self.get_device(afero_device.id)
        updated_keys = set()
        for state in afero_device.states:
            if state.functionClass == "available":
                if cur_item.available != state.value:
                    updated_keys.add("available")
                cur_item.available = state.value
            elif state.functionClass == "alarm-state":
                if cur_item.alarm_state.mode != state.value:
                    updated_keys.add(state.functionClass)
                cur_item.alarm_state.mode = state.value
            elif (
                (update_key := await self.update_sensor(state, cur_item))
                or (update_key := await self.update_number(state, cur_item))
                or (update_key := await self.update_select(state, cur_item))
            ):
                updated_keys.add(update_key)
            elif state.functionClass == "siren-action":
                try:
                    result_code = state.value["security-siren-action"]["resultCode"]
                    command = state.value["security-siren-action"]["command"]
                except TypeError:
                    result_code = None
                    command = None
                if (
                    result_code != cur_item.siren_action.result_code
                    or command != cur_item.siren_action.command
                ):
                    cur_item.siren_action.result_code = result_code
                    cur_item.siren_action.command = command
                    updated_keys.add("siren-action")

        return updated_keys

    async def set_state(
        self,
        device_id: str,
        disarm_pin: int | None = None,
        command: int | None = None,
        numbers: dict[tuple[str, str | None], float] | None = None,
        selects: dict[tuple[str, str | None], str] | None = None,
    ) -> None:
        """Set supported feature(s) to Security System resource."""
        update_obj = SecuritySystemPut()
        force_mode = False
        try:
            cur_item = self.get_device(device_id)
        except DeviceNotFound:
            self._logger.info("Unable to find device %s", device_id)
            return
        if command is not None:
            force_mode = True
            if command != 5:
                await self.validate_arm_state(device_id, command)
            update_obj.siren_action = features.SecuritySensorSirenFeature(
                result_code=0,
                command=command,
            )
        if disarm_pin is not None:
            force_mode = True
            update_obj.disarm_pin = features.SecuritySystemDisarmPin(pin=disarm_pin)
        if numbers:
            for key, val in numbers.items():
                if key not in cur_item.numbers:
                    continue
                update_obj.numbers[key] = features.NumbersFeature(
                    value=val,
                    min=cur_item.numbers[key].min,
                    max=cur_item.numbers[key].max,
                    step=cur_item.numbers[key].step,
                    name=cur_item.numbers[key].name,
                    unit=cur_item.numbers[key].unit,
                )
        if selects:
            for key, val in selects.items():
                if key not in cur_item.selects:
                    continue
                update_obj.selects[key] = features.SelectFeature(
                    selected=val,
                    selects=cur_item.selects[key].selects,
                    name=cur_item.selects[key].name,
                )
        await self.update(
            device_id, obj_in=update_obj, send_duplicate_states=force_mode
        )
        # Ensure the correct pin was used
        if disarm_pin:
            await self.validate_disarm_pin(device_id)
        elif command:
            await self.refresh_alarm_state(device_id)

    async def refresh_alarm_state(self, device_id: str) -> None:
        """Refresh the alarm state after alarm state change command."""
        task = asyncio.create_task(asyncio.sleep(UPDATE_TIME))
        await task
        states = await self._bridge.fetch_device_states(device_id)
        device = self._bridge.get_afero_device(device_id)
        device.states = states
        await self._bridge.events.generate_events_from_update(device)

    async def validate_disarm_pin(self, device_id: str) -> None:
        """Ensure the system has switched to disarmed."""
        # alarm-state does not update immediately, so wait and recheck
        task = asyncio.create_task(asyncio.sleep(UPDATE_TIME))
        await task
        states = await self._bridge.fetch_device_states(device_id)
        for state in states:
            if state.functionClass != "alarm-state":
                continue
            if state.value != "disarmed":
                raise SecuritySystemError("Disarm PIN was not accepted")
        device = self._bridge.get_afero_device(device_id)
        device.states = states
        await self._bridge.events.generate_events_from_update(device)

    async def validate_arm_state(self, device_id: str, arm_type: int) -> bool:
        """Ensure the system can arm."""
        dev = self._bridge.get_afero_device(device_id)
        sensors_with_issues = []
        num_sensors = 0
        for child in dev.children:
            controller = self._bridge.get_device_controller(child)
            if type(controller).__name__ != "SecuritySystemSensorController":
                continue
            sensor = controller[child]
            if (
                # Generic Bypass
                sensor.selects.get(("bypassType", None)).selected in ["Manual", "On"]
                or
                # Arm Away Bypass
                (
                    arm_type == 2
                    and sensor.selects.get(("triggerType", None)).selected
                    not in ["Away", "Home/Away"]
                )
                or
                # Arm Home Bypass
                (
                    arm_type == 4
                    and sensor.selects.get(("triggerType", None)).selected
                    not in ["Home", "Home/Away"]
                )
            ):
                self._logger.debug("Bypassing sensor %s", sensor.id)
                continue
            if sensor.available is False:
                sensors_with_issues.append(
                    f"{sensor.device_information.name} (Unavailable)"
                )
            if sensor.binary_sensors.get("triggered|None").current_value == "On":
                sensors_with_issues.append(
                    f"{sensor.device_information.name} (Triggered)"
                )
            if sensor.binary_sensors.get("tampered|None").current_value == "On":
                sensors_with_issues.append(
                    f"{sensor.device_information.name} (Tampered)"
                )
            num_sensors += 1
        if sensors_with_issues:
            raise SecuritySystemError(
                f"Sensors are open or unavailable: {', '.join(sensors_with_issues)}"
            )
        if num_sensors == 0:
            raise SecuritySystemError(
                "No sensors are configured for the requested mode."
            )
        return True



"""Controller that holds top-level devices."""

from aioafero.device import AferoDevice
from aioafero.v1.models.device import Device
from aioafero.v1.models.resource import DeviceInformation, ResourceTypes

from .base import AferoBinarySensor, AferoSensor, BaseResourcesController
from .event import AferoEvent, EventType


class DeviceController(BaseResourcesController[Device]):
    """Controller that identifies top-level components."""

    ITEM_TYPE_ID = ResourceTypes.DEVICE
    ITEM_TYPES = []
    ITEM_CLS = Device
    # Sensors map functionClass -> Unit
    ITEM_SENSORS: dict[str, str] = {
        "battery-level": "%",
        "wifi-rssi": "dB",
    }
    # Binary sensors map key -> alerting value
    ITEM_BINARY_SENSORS: dict[str, str] = {
        "error": "alerting",
    }

    def __init__(self, *args, **kwargs):
        """Initialize instance."""
        super().__init__(*args, **kwargs)
        self._known_parents: dict[str, str] = {}

    async def initialize_elem(self, afero_device: AferoDevice) -> Device:
        """Initialize the element.

        :param afero_device: Afero Device that contains the updated states

        :return: Newly initialized resource
        """
        available: bool = False
        sensors: dict[str, AferoSensor] = {}
        binary_sensors: dict[str, AferoBinarySensor] = {}
        wifi_mac: str | None = None
        ble_mac: str | None = None
        for state in afero_device.states:
            if state.functionClass == "available":
                available = state.value
            elif sensor := await self.initialize_sensor(state, afero_device.id):
                if isinstance(sensor, AferoBinarySensor):
                    binary_sensors[sensor.id] = sensor
                else:
                    sensors[sensor.id] = sensor
            elif state.functionClass == "wifi-mac-address":
                wifi_mac = state.value
            elif state.functionClass == "ble-mac-address":
                ble_mac = state.value

        self._items[afero_device.id] = Device(
            _id=afero_device.id,
            available=available,
            sensors=sensors,
            binary_sensors=binary_sensors,
            device_information=DeviceInformation(
                device_class=afero_device.device_class,
                default_image=afero_device.default_image,
                default_name=afero_device.default_name,
                manufacturer=afero_device.manufacturerName,
                model=afero_device.model,
                name=afero_device.friendly_name,
                parent_id=afero_device.device_id,
                wifi_mac=wifi_mac,
                ble_mac=ble_mac,
                version_data=getattr(afero_device, "version_data", {}),
                children=afero_device.children,
                functions=afero_device.functions,
            ),
        )
        return self._items[afero_device.id]

    async def initialize(self) -> None:
        """Initialize controller by fetching all items for this resource type from bridge."""
        if self._initialized:
            return
        # Subscribe to polled data to find all top-level devices
        self._bridge.events.subscribe(
            self._process_polled_devices,
            event_filter=EventType.POLLED_DEVICES,
        )
        # Subscribe to updates for existing devices
        self._bridge.events.subscribe(
            self._process_update_response,
            event_filter=EventType.RESOURCE_UPDATE_RESPONSE,
        )
        self._initialized = True

    async def _process_update_response(
        self, evt_type: EventType, evt_data: AferoEvent | None
    ) -> None:
        """Process the response of an update."""
        dev = evt_data["device"]
        if evt_data["device"].device_id in self._known_parents:
            evt = AferoEvent(
                type=EventType.RESOURCE_UPDATED,
                device_id=dev.id,
                device=dev,
            )
            await self._handle_event(evt["type"], evt)

    async def _process_polled_devices(
        self, evt_type: EventType, evt_data: AferoEvent | None
    ) -> None:
        """Find all top-level devices within the payload."""
        devices: list[AferoDevice] = evt_data["polled_devices"]
        parent_devices: list[AferoDevice] = self.get_filtered_devices(devices)
        processed: set[str] = set()
        for parent_device in parent_devices:
            evt = AferoEvent(
                type=EventType.RESOURCE_ADDED,
                device_id=parent_device.id,
                device=parent_device,
            )
            if parent_device.device_id not in self._known_parents:
                self._known_parents[parent_device.device_id] = parent_device.id
            else:
                evt["type"] = EventType.RESOURCE_UPDATED
            processed.add(parent_device.device_id)
            await self._handle_event(evt["type"], evt)
        for known_id in list(self._known_parents):
            if known_id not in processed:
                device_id = self._known_parents.pop(known_id)
                self._logger.info("Device %s was not polled. Removing", known_id)
                evt = AferoEvent(
                    type=EventType.RESOURCE_DELETED,
                    device_id=device_id,
                )
                await self._handle_event(evt["type"], evt)

    def get_filtered_devices(self, devices: list[AferoDevice]) -> list[AferoDevice]:
        """Find parent devices."""
        parents: dict = {}
        potential_parents: dict = {}
        for device in devices:
            if device.children:
                parents[device.device_id] = device
            elif device.device_id not in parents and (
                device.device_id not in parents
                and device.device_id not in potential_parents
            ):
                potential_parents[device.device_id] = device
            else:
                self._logger.debug("skipping %s as its tracked", device.device_id)
        for potential_parent in potential_parents.values():
            if potential_parent.device_id not in parents:
                parents[potential_parent.device_id] = potential_parent
        return list(parents.values())

    async def update_elem(self, afero_device: AferoDevice) -> set:
        """Update the Device with the latest API data.

        :param afero_device: Afero Device that contains the updated states

        :return: States that have been modified
        """
        cur_item = self.get_device(afero_device.id)
        updated_keys = set()
        for state in afero_device.states:
            if state.functionClass == "available":
                if cur_item.available != state.value:
                    cur_item.available = state.value
                    updated_keys.add(state.functionClass)
            elif update_key := await self.update_sensor(state, cur_item):
                updated_keys.add(update_key)
        return updated_keys



import pytest

from aioafero.v1.models import SecuritySystemSensor, features, DeviceInformation
from aioafero.v1.models.sensor import AferoBinarySensor, AferoSensor


@pytest.fixture
def populated_entity():
    return SecuritySystemSensor(
        _id="7f4e4c01-e799-45c5-9b1a-385433a78edc-sensor-2",
        split_identifier="sensor",
        available=True,
        selects={
            ("sensor-2 ", "chirpMode"): features.SelectFeature(
                selected="On",
                selects={"On", "Off"},
                name="Chirp Mode",
            ),
        },
        binary_sensors={
            "tampered": AferoBinarySensor(
                id="tampered",
                owner="7f4e4c01-e799-45c5-9b1a-385433a78edc-sensor-2",
                instance="tampered",
                current_value=1,
                _error=1,
            )
        },
        sensors={
            "batteryLevel": AferoSensor(
                id="sensor-state",
                owner="7f4e4c01-e799-45c5-9b1a-385433a78edc-sensor-2",
                value=100,
                unit="%",
                instance=None,
            ),
        },
        device_information=DeviceInformation(
            functions=[
            {
                "functionClass": "preset",
                "functionInstance": "preset-1",
                "value": "on",
                "lastUpdateTime": 0,
            }
        ]
        )
    )


@pytest.fixture
def empty_entity():
    return SecuritySystemSensor(
        _id="7f4e4c01-e799-45c5-9b1a-385433a78edc-sensor-2",
        split_identifier="sensor",
        available=True,
        selects={},
        binary_sensors={},
        sensors={},
        device_information=DeviceInformation(
            functions=[
            {
                "functionClass": "preset",
                "functionInstance": "preset-1",
                "value": "on",
                "lastUpdateTime": 0,
            }
        ]
        )
    )


def test_init(populated_entity):
    assert populated_entity.id == "7f4e4c01-e799-45c5-9b1a-385433a78edc-sensor-2"
    assert populated_entity.available is True
    assert populated_entity.instances == {"preset": "preset-1"}
    assert populated_entity.update_id == "7f4e4c01-e799-45c5-9b1a-385433a78edc"
    assert populated_entity.instance == 2



"""Controller holding and managing Afero IoT resources of type `fan`."""

from aioafero.device import AferoDevice, get_function_from_device
from aioafero.util import ordered_list_item_to_percentage
from aioafero.v1.models import features
from aioafero.v1.models.fan import Fan, FanPut
from aioafero.v1.models.resource import DeviceInformation, ResourceTypes

from .base import AferoBinarySensor, AferoSensor, BaseResourcesController

KNOWN_PRESETS = {"comfort-breeze"}


class FanController(BaseResourcesController[Fan]):
    """Controller holding and managing Afero IoT resources of type `fan`."""

    ITEM_TYPE_ID = ResourceTypes.DEVICE
    ITEM_TYPES = [ResourceTypes.FAN]
    ITEM_CLS = Fan
    ITEM_MAPPING = {
        "on": "power",
        "speed": "fan-speed",
        "direction": "fan-reverse",
    }

    async def turn_on(self, device_id: str) -> None:
        """Turn on the fan."""
        await self.set_state(device_id, on=True)

    async def turn_off(self, device_id: str) -> None:
        """Turn off the fan."""
        await self.set_state(device_id, on=False)

    async def set_speed(self, device_id: str, speed: int) -> None:
        """Set the speed of the fan, as a percentage."""
        await self.set_state(device_id, on=True, speed=speed)

    async def set_direction(self, device_id: str, forward: bool) -> None:
        """Set the direction of the fan to forward."""
        cur_item = self.get_device(device_id)
        if not cur_item.is_on:
            # Thanks Hubspace for this one! Additionally, turning it on and setting
            # direction at the same time does not work as expected
            self._logger.info("Fan is not running so direction will not be set")
        await self.set_state(device_id, forward=forward)

    async def set_preset(self, device_id: str, preset: bool) -> None:
        """Set the preset of the fan."""
        await self.set_state(device_id, on=True, preset=preset)

    async def initialize_elem(self, afero_device: AferoDevice) -> Fan:
        """Initialize the element.

        :param afero_device: Afero Device that contains the updated states

        :return: Newly initialized resource
        """
        available: bool = False
        on: features.OnFeature | None = None
        speed: features.SpeedFeature | None = None
        direction: features.DirectionFeature | None = None
        preset: features.PresetFeature | None = None
        sensors: dict[str, AferoSensor] = {}
        binary_sensors: dict[str, AferoBinarySensor] = {}
        for state in afero_device.states:
            if state.functionClass == "power":
                on = features.OnFeature(on=state.value == "on")
            elif state.functionClass == "fan-speed":
                speeds = get_function_from_device(
                    afero_device.functions, state.functionClass, state.functionInstance
                )
                tmp_speed = set()
                for value in speeds["values"]:
                    if not value["name"].endswith("-000"):
                        tmp_speed.add(value["name"])
                speeds = sorted(tmp_speed)
                percentage = ordered_list_item_to_percentage(speeds, state.value)
                speed = features.SpeedFeature(speed=percentage, speeds=speeds)
            elif state.functionClass == "fan-reverse":
                direction = features.DirectionFeature(forward=state.value == "forward")
            elif (
                state.functionClass == "toggle"
                and state.functionInstance in KNOWN_PRESETS
            ):
                # I have only seen fans with a single preset
                preset = features.PresetFeature(
                    enabled=state.value == "enabled",
                    func_class=state.functionClass,
                    func_instance=state.functionInstance,
                )
            elif state.functionClass == "available":
                available = state.value

        self._items[afero_device.id] = Fan(
            _id=afero_device.id,
            available=available,
            sensors=sensors,
            binary_sensors=binary_sensors,
            device_information=DeviceInformation(
                device_class=afero_device.device_class,
                default_image=afero_device.default_image,
                default_name=afero_device.default_name,
                manufacturer=afero_device.manufacturerName,
                model=afero_device.model,
                name=afero_device.friendly_name,
                parent_id=afero_device.device_id,
                children=afero_device.children,
                functions=afero_device.functions,
            ),
            on=on,
            speed=speed,
            direction=direction,
            preset=preset,
        )
        return self._items[afero_device.id]

    async def update_elem(self, afero_device: AferoDevice) -> set:
        """Update the Fan with the latest API data.

        :param afero_device: Afero Device that contains the updated states

        :return: States that have been modified
        """
        updated_keys = set()
        cur_item = self.get_device(afero_device.id)
        for state in afero_device.states:
            if state.functionClass == "power":
                new_val = state.value == "on"
                if cur_item.on.on != new_val:
                    cur_item.on.on = new_val
                    updated_keys.add("on")
            elif state.functionClass == "fan-speed":
                new_val = ordered_list_item_to_percentage(
                    cur_item.speed.speeds, state.value
                )
                if cur_item.speed.speed != new_val:
                    cur_item.speed.speed = new_val
                    updated_keys.add("speed")
            elif state.functionClass == "fan-reverse":
                new_val = state.value == "forward"
                if cur_item.direction.forward != new_val:
                    cur_item.direction.forward = new_val
                    updated_keys.add("direction")
            elif (
                state.functionClass == "toggle"
                and state.functionInstance in KNOWN_PRESETS
            ):
                new_val = state.value == "enabled"
                if cur_item.preset.enabled != new_val:
                    cur_item.preset.enabled = new_val
                    updated_keys.add("preset")
            elif state.functionClass == "available":
                if cur_item.available != state.value:
                    cur_item.available = state.value
                    updated_keys.add("available")

        return updated_keys

    async def set_state(
        self,
        device_id: str,
        on: bool | None = None,
        speed: int | None = None,
        forward: bool | None = None,
        preset: bool | None = None,
    ) -> None:
        """Set supported feature(s) to fan resource."""
        update_obj = FanPut()
        cur_item = self.get_device(device_id)
        if on is not None:
            update_obj.on = features.OnFeature(on=on)
        if speed is not None and cur_item.speed is not None:
            if speed == 0:
                update_obj.on = features.OnFeature(on=False)
            else:
                update_obj.speed = features.SpeedFeature(
                    speed=speed, speeds=cur_item.speed.speeds
                )
        if preset is not None and cur_item.preset is not None:
            update_obj.preset = features.PresetFeature(
                enabled=preset,
                func_class=cur_item.preset.func_class,
                func_instance=cur_item.preset.func_instance,
            )
        if forward is not None and cur_item.direction is not None:
            update_obj.direction = features.DirectionFeature(forward=forward)
        await self.update(device_id, obj_in=update_obj)



"""Representation of an Afero Fan and its corresponding updates."""

from dataclasses import dataclass

from aioafero.v1.models import features

from .resource import ResourceTypes
from .standard_mixin import StandardMixin


@dataclass(kw_only=True)
class Fan(StandardMixin):
    """Representation of an Afero Fan."""

    type: ResourceTypes = ResourceTypes.FAN

    on: features.OnFeature | None = None
    speed: features.SpeedFeature | None = None
    direction: features.DirectionFeature | None = None
    preset: features.PresetFeature | None = None

    @property
    def supports_direction(self):
        """Determine if you can change the direction of the fan."""
        return self.direction is not None

    @property
    def supports_on(self):
        """Determine if you can turn the fan on or off."""
        return self.on is not None

    @property
    def supports_presets(self):
        """Determine if presets are supported by this fan."""
        return self.preset is not None

    @property
    def supports_speed(self):
        """Determine if a speed feature is supported by this fan."""
        return self.speed is not None

    @property
    def is_on(self) -> bool:
        """Return bool if fan is currently powered on."""
        if self.on:
            return self.on.on
        return False

    @property
    def current_direction(self) -> bool:
        """Return if the direction is forward."""
        if self.direction:
            return self.direction.forward
        return False

    @property
    def current_speed(self) -> int:
        """Current speed of the fan, as a percentage."""
        if self.speed:
            return self.speed.speed
        return 0

    @property
    def current_preset(self) -> str | None:
        """Current fan preset."""
        if self.preset and self.preset.enabled:
            return self.preset.func_instance
        return None


@dataclass
class FanPut:
    """States that can be updated for a Fan."""

    on: features.OnFeature | None = None
    speed: features.SpeedFeature | None = None
    direction: features.DirectionFeature | None = None
    preset: features.PresetFeature | None = None



"""Types used by the bridge and supporting elements."""

from enum import Enum


class EventType(Enum):
    """Enum with possible Events."""

    RESOURCE_ADDED = "add"
    RESOURCE_UPDATED = "update"
    RESOURCE_DELETED = "delete"
    RESOURCE_VERSION = "version"
    RESOURCE_UPDATE_RESPONSE = "update_response"
    CONNECTED = "connected"
    DISCONNECTED = "disconnected"
    RECONNECTED = "reconnected"
    INVALID_AUTH = "invalid_auth"
    POLLED_DATA = "polled_data"
    POLLED_DEVICES = "polled_devices"


class TemperatureUnit(Enum):
    """Temperature unit enum."""

    CELSIUS = "celsius"
    FAHRENHEIT = "fahrenheit"



"""Representation of an Afero Portable AC and its corresponding updates."""

from dataclasses import dataclass, field

from aioafero.v1.models import features

from .hvac_mixin import HVACMixin
from .resource import ResourceTypes
from .standard_mixin import StandardMixin

mode_checks = {
    "auto-cool": "cool",
}


@dataclass(kw_only=True)
class PortableAC(StandardMixin, HVACMixin):
    """Representation of an Afero Portable AC."""

    type: ResourceTypes = ResourceTypes.PORTABLE_AC

    numbers: dict[tuple[str, str | None], features.NumbersFeature] | None
    selects: dict[tuple[str, str | None], features.SelectFeature] | None

    @property
    def supports_fan_mode(self) -> bool:
        """Can enable fan-only mode."""
        return False

    def get_mode_to_check(self) -> str | None:
        """Determine the current mode of the thermostat."""
        if not self.hvac_mode:
            return None
        checked_mode = mode_checks.get(self.hvac_mode.mode, self.hvac_mode.mode)
        checked_prev_mode = mode_checks.get(
            self.hvac_mode.previous_mode, self.hvac_mode.previous_mode
        )
        if checked_mode in [
            "cool",
            "heat",
        ]:
            return checked_mode
        if checked_mode in ["dehumidify"]:
            return checked_prev_mode
        return None

    @property
    def supports_temperature_range(self) -> bool:
        """Range which the thermostat will heat / cool."""
        return False


@dataclass
class PortableACPut:
    """States that can be updated for a Portable AC."""

    # This feels wrong but based on data dumps, setting timer increases the
    # current temperature by 1 to turn it on
    current_temperature: features.CurrentTemperatureFeature | None = None
    hvac_mode: features.HVACModeFeature | None = None
    target_temperature_cooling: features.TargetTemperatureFeature | None = None
    numbers: dict[tuple[str, str | None], features.NumbersFeature] | None = field(
        default_factory=dict, repr=False, init=False
    )
    selects: dict[tuple[str, str | None], features.SelectFeature] | None = field(
        default_factory=dict, repr=False, init=False
    )



import json
import os

import pytest

from aioafero import device
from pathlib import Path

current_path = os.path.dirname(os.path.realpath(__file__))


with open(os.path.join(current_path, "v1", "data", "device_lock.json")) as fh:
    device_lock_response = json.load(fh)
    lock_dev = device.get_afero_device(device_lock_response[0])


@pytest.mark.parametrize(
    "afero_device,expected",
    [
        # Everything is set correctly
        (
            {
                "id": "id",
                "device_id": "device_id",
                "model": "model",
                "device_class": "device_class",
                "default_name": "default_name",
                "default_image": "default_image",
                "friendly_name": "friendly_name",
                "functions": ["functions!"],
                "states": [],
            },
            device.AferoDevice(
                id="id", device_id="device_id", model="model", device_class="device_class", default_name="default_name", default_image="default_image", friendly_name="friendly_name", functions=["functions!"], states=[]
            ),
        ),
        # DriskolFan
        (
            {
                "id": "id",
                "device_id": "device_id",
                "model": "",
                "device_class": "fan",
                "default_name": "default_name",
                "default_image": "ceiling-fan-snyder-park-icon",
                "friendly_name": "friendly_name",
                "functions": ["functions!"],
                "states": [],
            },
            device.AferoDevice(
                id="id", device_id="device_id", model="Driskol", device_class="fan", default_name="default_name", default_image="ceiling-fan-snyder-park-icon", friendly_name="friendly_name", functions=["functions!"], states=[]
            ),
        ),
        # VinwoodFan
        (
            {
                "id": "id",
                "device_id": "device_id",
                "model": "",
                "device_class": "fan",
                "default_name": "default_name",
                "default_image": "ceiling-fan-vinings-icon",
                "friendly_name": "friendly_name",
                "functions": ["functions!"],
                "states": [],
            },
            device.AferoDevice(
                id="id", device_id="device_id", model="Vinwood", device_class="fan", default_name="default_name", default_image="ceiling-fan-vinings-icon", friendly_name="friendly_name", functions=["functions!"], states=[]
            ),
        ),
        # ZandraFan
        (
            {
                "id": "id",
                "device_id": "device_id",
                "model": "TBD",
                "device_class": "fan",
                "default_name": "default_name",
                "default_image": "ceiling-fan-chandra-icon",
                "friendly_name": "friendly_name",
                "functions": ["functions!"],
                "states": [],
            },
            device.AferoDevice(
                id="id", device_id="device_id", model="Zandra", device_class="fan", default_name="default_name", default_image="ceiling-fan-chandra-icon", friendly_name="friendly_name", functions=["functions!"], states=[]
            ),
        ),
        # NevaliFan
        (
            {
                "id": "id",
                "device_id": "device_id",
                "model": "TBD",
                "device_class": "fan",
                "default_name": "default_name",
                "default_image": "ceiling-fan-ac-cct-dardanus-icon",
                "friendly_name": "friendly_name",
                "functions": ["functions!"],
                "states": [],
            },
            device.AferoDevice(
                id="id", device_id="device_id", model="Nevali", device_class="fan", default_name="default_name", default_image="ceiling-fan-ac-cct-dardanus-icon", friendly_name="friendly_name", functions=["functions!"], states=[]
            ),
        ),
        # TagerFan
        (
            {
                "id": "id",
                "device_id": "device_id",
                "model": "",
                "device_class": "fan",
                "default_name": "default_name",
                "default_image": "ceiling-fan-slender-icon",
                "friendly_name": "friendly_name",
                "functions": ["functions!"],
                "states": [],
            },
            device.AferoDevice(
                id="id", device_id="device_id", model="Tager", device_class="fan", default_name="default_name", default_image="ceiling-fan-slender-icon", friendly_name="friendly_name", functions=["functions!"], states=[]
            ),
        ),
        # Dimmer switch
        (
            {
                "id": "id",
                "device_id": "device_id",
                "model": "dimmer",
                "device_class": "switch",
                "default_name": "default_name",
                "default_image": "ceiling-fan-slender-icon",
                "friendly_name": "friendly_name",
                "functions": ["functions!"],
                "states": [
                    device.AferoState(
                        functionClass="brightness", functionInstance=None, lastUpdateTime=1668551478232, value=40
                    )
                ],
            },
            device.AferoDevice(
                id="id", device_id="device_id", model="dimmer", device_class="light", default_name="default_name", default_image="ceiling-fan-slender-icon", friendly_name="friendly_name", functions=["functions!"], states=[
                        device.AferoState(
                            functionClass="brightness", functionInstance=None, lastUpdateTime=1668551478232, value=40
                        )
                    ]
            ),
        ),
        # Glass door
        (
            {
                "id": "id",
                "device_id": "device_id",
                "model": "glass-door",
                "device_class": "glass-door",
                "default_name": "default_name",
                "default_image": "glass-door-icon",
                "friendly_name": "friendly_name",
                "functions": ["functions!"],
                "states": [
                    device.AferoState(
                        functionClass="power", functionInstance=None, lastUpdateTime=1668551478232, value="off"
                    )
                ],
            },
            device.AferoDevice(
                id="id", device_id="device_id", model="glass-door", device_class="switch", default_name="default_name", default_image="glass-door-icon", friendly_name="friendly_name", functions=["functions!"], states=[
                        device.AferoState(
                            functionClass="power", functionInstance=None, lastUpdateTime=1668551478232, value="off"
                        )
                    ], manufacturerName="Feather River Doors"
            ),
        ),
        # Exhaust fan
        (
            {
                "id": "id",
                "device_id": "device_id",
                "model": "glass-door",
                "device_class": "exhaust-fan",
                "default_name": "default_name",
                "default_image": "fan-exhaust-icon",
                "friendly_name": "friendly_name",
                "functions": ["functions!"],
                "states": [
                    device.AferoState(
                        functionClass="power", functionInstance=None, lastUpdateTime=1668551478232, value="off"
                    )
                ],
            },
            device.AferoDevice(
                id="id", device_id="device_id", model="BF1112", device_class="exhaust-fan", default_name="default_name", default_image="fan-exhaust-icon", friendly_name="friendly_name", functions=["functions!"], states=[
                        device.AferoState(
                            functionClass="power", functionInstance=None, lastUpdateTime=1668551478232, value="off"
                        )
                    ], manufacturerName=None
            ),
        ),
        # 12A19060WRGBWH2
        (
            {
                "id": "id",
                "device_id": "device_id",
                "model": "glass-door",
                "device_class": "light",
                "default_name": "default_name",
                "default_image": "a19-e26-color-cct-60w-smd-frosted-icon",
                "friendly_name": "friendly_name",
                "functions": ["functions!"],
                "states": [
                    device.AferoState(
                        functionClass="power", functionInstance=None, lastUpdateTime=1668551478232, value="off"
                    )
                ],
            },
            device.AferoDevice(
                id="id", device_id="device_id", model="12A19060WRGBWH2", device_class="light", default_name="default_name", default_image="a19-e26-color-cct-60w-smd-frosted-icon", friendly_name="friendly_name", functions=["functions!"], states=[
                        device.AferoState(
                            functionClass="power", functionInstance=None, lastUpdateTime=1668551478232, value="off"
                        )
                    ], manufacturerName=None
            ),
        ),
        (
            {
                "id": "id",
                "device_id": "device_id",
                "model": "glass-door",
                "device_class": "light",
                "default_name": "default_name",
                "default_image": "slide-dimmer-icon",
                "friendly_name": "friendly_name",
                "functions": ["functions!"],
                "states": [
                    device.AferoState(
                        functionClass="power", functionInstance=None, lastUpdateTime=1668551478232, value="off"
                    )
                ],
            },
            device.AferoDevice(
                id="id", device_id="device_id", model="HPDA110NWBP", device_class="light", default_name="default_name", default_image="slide-dimmer-icon", friendly_name="friendly_name", functions=["functions!"], states=[
                        device.AferoState(
                            functionClass="power", functionInstance=None, lastUpdateTime=1668551478232, value="off"
                        )
                    ], manufacturerName=None
            ),
        ),
        (
            {
                "id": "id",
                "device_id": "device_id",
                "model": "TBD",
                "device_class": "switch",
                "default_name": "default_name",
                "default_image": "smart-switch-icon",
                "friendly_name": "friendly_name",
                "functions": ["functions!"],
                "states": [
                    device.AferoState(
                        functionClass="power", functionInstance=None, lastUpdateTime=1668551478232, value="off"
                    )
                ],
            },
            device.AferoDevice(
                id="id", device_id="device_id", model="HPSA11CWB", device_class="switch", default_name="default_name", default_image="smart-switch-icon", friendly_name="friendly_name", functions=["functions!"], states=[
                        device.AferoState(
                            functionClass="power", functionInstance=None, lastUpdateTime=1668551478232, value="off"
                        )
                    ], manufacturerName=None
            ),
        ),
        # LCN3002LM-01 WH
        (
            {
                "id": "id",
                "device_id": "device_id",
                "model": "model",
                "device_class": "light",
                "default_name": "default_name",
                "default_image": "bright-edgelit-flushmount-light-icon",
                "friendly_name": "friendly_name",
                "functions": ["functions!"],
                "states": [],
            },
            device.AferoDevice(
                id="id", device_id="device_id", manufacturerName="Commercial-Electric", model="LCN3002LM-01 WH", device_class="light", default_name="default_name", default_image="bright-edgelit-flushmount-light-icon", friendly_name="friendly_name", functions=["functions!"], states=[]
            ),
        ),
    ],
)
def test_AferoDevice(afero_device, expected):
    assert device.AferoDevice(**afero_device) == expected


@pytest.mark.parametrize(
    "afero_device,expected_attrs",
    [
        # Validate when values are missing
        (
            {},
            {
                "id": None,
                "device_id": None,
                "model": None,
                "device_class": None,
                "default_name": None,
                "default_image": None,
                "friendly_name": None,
                "functions": [],
            },
        ),
        # Ensure values are properly parsed
        (
            device_lock_response[0],
            {
                "id": "5a5d5e04-a6ad-47c0-b9f4-b9fe5c049ef4",
                "device_id": "0123f95ec14bdb23",
                "model": "Keypad Deadbolt Lock",
                "device_class": "door-lock",
                "default_name": "Keypad Deadbolt Lock",
                "default_image": "keypad-deadbolt-lock-icon",
                "friendly_name": "Friendly Name 2",
                "functions": device_lock_response[0]["description"]["functions"],
            },
        ),
    ],
)
def test_get_afero_device(afero_device, expected_attrs):
    dev = device.get_afero_device(afero_device)
    for key, val in expected_attrs.items():
        assert (
            getattr(dev, key) == val
        ), f"Key {key} did not match, {getattr(dev, key)} != {val}"


@pytest.mark.parametrize(
    "data,expected_attrs",
    [
        # verify defaults
        (
            {
                "functionClass": "class",
                "value": "beans",
            },
            {
                "functionClass": "class",
                "value": "beans",
                "functionInstance": None,
                "lastUpdateTime": None,
            },
        ),
        # verify defaults
        (
            {
                "functionClass": "class",
                "value": "beans",
                "functionInstance": "beans",
                "lastUpdateTime": 4,
            },
            {
                "functionClass": "class",
                "value": "beans",
                "functionInstance": "beans",
                "lastUpdateTime": 4,
            },
        ),
    ],
)
def test_AferoState(data, expected_attrs):
    elem = device.AferoState(**data)
    for key, val in expected_attrs.items():
        assert getattr(elem, key) == val


def test_AferoDevice_hash():
    dev = device.get_afero_device(device_lock_response[0])
    hash_check = {dev: True}
    assert dev in hash_check


@pytest.mark.parametrize(
    "functions, func_class, func_instance, expected",
    [
        ([], "cool", "beans", None),
        (lock_dev.functions, "lock-pin", None, None),
        (lock_dev.functions, "lock-pin", "lock-pin-9", lock_dev.functions[19]),
    ],
)
def test_get_function_from_device(functions, func_class, func_instance, expected):
    assert (
        device.get_function_from_device(functions, func_class, func_instance)
        == expected
    )



@pytest.mark.parametrize(
        (("capability"), ("expected")), [
            (
                {
                    "functionClass": "sensor-state",
                    "functionInstance": "sensor-1",
                    "type": "object",
                    "schedulable": False,
                    "name": "Aaaaa",
                    "locale": "en_US"
                },
                device.AferoCapability(
                    functionClass="sensor-state",
                    type="object",
                    schedulable=False,
                    functionInstance="sensor-1",
                    _opts={
                        "name": "Aaaaa",
                        "locale": "en_US"
                    }
                )
            ),
            (
                {
                    "functionClass": "siren-action",
                    "type": "object",
                    "schedulable": False
                },
                device.AferoCapability(
                    functionClass="siren-action",
                    type="object",
                    schedulable=False,
                )
            )
        ]
)
def test_transform_capability(capability, expected):
    actual = device.transform_capability(capability)
    assert actual == expected
    assert actual.options == expected._opts


def test_get_afero_device():
    path_to_file: Path = Path(__file__).parent / "v1" / "data" / "device-with-capabilities.json"
    with path_to_file.open() as fh:
        data = json.load(fh)
    dev = device.get_afero_device(data[0])
    assert len(dev.capabilities) == 102


capabilities = [
    device.AferoCapability(
        **{
            "functionClass": "temperature",
            "type": "numeric",
            "schedulable": True,
            "functionInstance": "cooling-target",
            "_opts": {
                "range": {
                    "min": 60,
                    "max": 86,
                    "step": 1
                }
            }
        }
    )
]

@pytest.mark.parametrize(
    ("capabilities", "func_class", "func_instance", "expected"),
    [
        (capabilities, "temperature", "cooling-target", capabilities[0]),
        (capabilities, "temperature", "heating-target", None),
        (capabilities, "none-instance", "heating-target", None),
    ],
)
def test_get_capability_from_device(capabilities, func_class, func_instance, expected):
    assert device.get_capability_from_device(capabilities, func_class, func_instance) == expected


def test_capability_raw_dump():
    cap = capabilities[0]
    assert cap.raw_dump() == {
        "functionClass": "temperature",
        "type": "numeric",
        "schedulable": True,
        "functionInstance": "cooling-target",
        "range": {
            "min": 60,
            "max": 86,
            "step": 1
        }
    }



import pytest

from aioafero.v1.models import features, DeviceInformation
from aioafero.v1.models.lock import Lock


@pytest.fixture
def populated_entity():
    return Lock(
        _id="entity-1",
        available=True,
        position=features.CurrentPositionFeature(
            position=features.CurrentPositionEnum.LOCKED
        ),
        device_information=DeviceInformation(
            functions=[
            {
                "functionClass": "preset",
                "functionInstance": "preset-1",
                "value": "on",
                "lastUpdateTime": 0,
            }
        ]
        )
    )


def test_init(populated_entity):
    assert populated_entity.id == "entity-1"


def test_get_instance(populated_entity):
    assert populated_entity.get_instance("preset") == "preset-1"



"""Controls Hubspace devices on v1 API."""

__all__ = [
    "AferoBridgeV1",
    "AferoController",
    "AferoModelResource",
    "BaseResourcesController",
    "DeviceController",
    "FanController",
    "LightController",
    "LockController",
    "PortableACController",
    "SecuritySystemController",
    "SecuritySystemKeypadController",
    "SecuritySystemSensorController",
    "SwitchController",
    "ThermostatController",
    "TokenData",
    "ValveController",
    "models",
]

import asyncio
from collections.abc import Callable, Generator
import contextlib
from contextlib import asynccontextmanager
import logging
from typing import Any

import aiohttp
from aiohttp import web_exceptions
from securelogging import LogRedactorMessage, add_secret

from aioafero.device import AferoDevice, AferoResource, AferoState
from aioafero.errors import (
    AferoError,
    DeviceNotFound,
    ExceededMaximumRetries,
    InvalidAuth,
)
from aioafero.types import TemperatureUnit

from . import models, v1_const
from .auth import AferoAuth, TokenData, passthrough
from .controllers.base import AferoBinarySensor, AferoSensor, BaseResourcesController
from .controllers.device import DeviceController
from .controllers.event import EventCallBackType, EventStream, EventType
from .controllers.exhaust_fan import ExhaustFanController
from .controllers.fan import FanController
from .controllers.light import LightController
from .controllers.lock import LockController
from .controllers.portable_ac import PortableACController
from .controllers.security_system import SecuritySystemController
from .controllers.security_system_keypad import SecuritySystemKeypadController
from .controllers.security_system_sensor import SecuritySystemSensorController
from .controllers.switch import SwitchController
from .controllers.thermostat import ThermostatController
from .controllers.valve import ValveController

type AferoModelResource = (
    models.Device
    | models.Fan
    | models.Light
    | models.Lock
    | models.Switch
    | models.Valve
    | models.Thermostat
    | AferoBinarySensor
    | AferoSensor
    | models.ExhaustFan
    | models.PortableAC
    | models.SecuritySystem
    | models.SecuritySystemSensor
)

type AferoController = (
    DeviceController
    | FanController
    | LightController
    | LockController
    | AferoSensor
    | SwitchController
    | ThermostatController
    | ValveController
    | ExhaustFanController
    | PortableACController
    | SecuritySystemController
    | SecuritySystemKeypadController
    | SecuritySystemSensorController
)


class AferoBridgeV1:
    """Controls Afero IoT devices on v1 API.

    This class serves as the main entry point for interacting with the Afero API.
    It handles authentication, device discovery, state management, and event handling.

    :param username: The username for the Afero-backed account (e.g., Hubspace).
    :param password: The password for the Afero-backed account.
    :param refresh_token: An optional refresh token to bypass initial login.
    :param session: An optional `aiohttp.ClientSession` to use for requests.
        If not provided, a new session will be created.
    :param polling_interval: The interval in seconds between polling the Afero API
        for device state updates. Defaults to 30.
    :param afero_client: The Afero client identifier (e.g., "hubspace", "myko").
        Defaults to "hubspace".
    :param hide_secrets: If True, sensitive information will be redacted from logs.
        Defaults to True.
    :param poll_version: If True, device version information will be polled periodically.
        Defaults to True.
    :param client_name: A name for the client to be used in the User-Agent header.
        Defaults to "aioafero".
    :param temperature_unit: The desired temperature unit for API responses.
        Defaults to `TemperatureUnit.CELSIUS`.

    """

    _web_session: aiohttp.ClientSession | None = None

    def __init__(
        self,
        username: str,
        password: str,
        refresh_token: str | None = None,
        session: aiohttp.ClientSession | None = None,
        polling_interval: int = 30,
        afero_client: str | None = "hubspace",
        hide_secrets: bool = True,
        poll_version: bool = True,
        client_name: str | None = "aioafero",
        temperature_unit: TemperatureUnit = TemperatureUnit.CELSIUS,
    ):
        """Initialize the AferoBridgeV1 instance."""
        if hide_secrets:
            self.secret_logger = LogRedactorMessage
        else:
            self.secret_logger = passthrough
        self._close_session: bool = session is None
        self._web_session: aiohttp.ClientSession = session
        self._account_id: str | None = None
        self._afero_client: str = afero_client
        self._auth = AferoAuth(
            self,
            username,
            password,
            refresh_token=refresh_token,
            afero_client=afero_client,
            hide_secrets=hide_secrets,
        )
        self.client_name = client_name
        self.temperature_unit = temperature_unit
        self.logger = logging.getLogger(f"{__package__}-{afero_client}[{username}]")
        if len(self.logger.handlers) == 0:
            self.logger.addHandler(logging.StreamHandler())
        self._known_devs: dict[str, BaseResourcesController] = {}
        self._known_afero_devices: dict[str, str] = {}
        # Known running tasks
        self._scheduled_tasks: list[asyncio.Task] = []
        self._adhoc_tasks: list[asyncio.Task] = []
        # Data Updater
        self._events: EventStream = EventStream(self, polling_interval, poll_version)
        # Data Controllers
        self._controllers: dict[str, BaseResourcesController] = {}
        self.add_controller("devices", DeviceController)
        self.add_controller("exhaust_fans", ExhaustFanController)
        self.add_controller("fans", FanController)
        self.add_controller("lights", LightController)
        self.add_controller("locks", LockController)
        self.add_controller("portable_acs", PortableACController)
        self.add_controller("security_systems", SecuritySystemController)
        self.add_controller("security_systems_keypads", SecuritySystemKeypadController)
        self.add_controller("security_systems_sensors", SecuritySystemSensorController)
        self.add_controller("switches", SwitchController)
        self.add_controller("thermostats", ThermostatController)
        self.add_controller("valves", ValveController)

    @property
    def refresh_token(self) -> str | None:
        """Get the current sessions refresh token."""
        return self._auth.refresh_token

    @property
    def events(self) -> EventStream:
        """Get the class that handles getting new data and notifying controllers."""
        return self._events

    @property
    def controllers(self) -> list:
        """Get a list of initialized controllers."""
        return [
            controller
            for controller in self._controllers.values()
            if controller.initialized
        ]

    @property
    def tracked_devices(self) -> set:
        """Get all tracked devices."""
        return set(self._known_devs.keys())

    async def otp_login(self, otp_code: str) -> None:
        """Perform OTP login with the provided code."""
        task = asyncio.create_task(self._auth.perform_otp_login(otp_code))
        self.add_job(task)
        await task
        return task.result()

    def add_device(
        self, device_id: str, controller: BaseResourcesController[AferoResource]
    ) -> None:
        """Add a device to the list of known devices and map it to its controller.

        :param device_id: The unique identifier of the device.
        :param controller: The controller instance responsible for the device.
        """
        self._known_devs[device_id] = controller

    def get_device_controller(self, device_id: str) -> BaseResourcesController:
        """Get the controller for a given device."""
        try:
            return self._known_devs[device_id]
        except KeyError as err:
            raise DeviceNotFound(f"Unable to find device {device_id}") from err

    def remove_device(self, device_id: str) -> None:
        """Remove a device from the list of known devices.

        :param device_id: The unique identifier of the device to remove.
        """
        with contextlib.suppress(KeyError):
            self._known_devs.pop(device_id)
        with contextlib.suppress(KeyError):
            self._known_afero_devices.pop(device_id)

    def add_afero_dev(self, device: AferoDevice, device_id: str | None = None) -> None:
        """Add a raw AferoDevice object to the internal cache.

        :param device: The `AferoDevice` object to cache.
        :param device_id: The ID to use for caching. If None, `device.id` is used.
        """
        if not device_id:
            device_id = device.id
        self._known_afero_devices[device_id] = device

    def get_afero_device(self, device_id: str) -> AferoDevice | None:
        """Get the raw AferoDevice object for a given ID.

        :param device_id: The unique identifier of the device.

        :return: The `AferoDevice` object if found, otherwise raises `DeviceNotFound`.
        :raises DeviceNotFound: If the device with the given ID is not found.
        """
        try:
            return self._known_afero_devices[device_id]
        except KeyError as err:
            raise DeviceNotFound(f"Unable to find device for {device_id}") from err

    @property
    def account_id(self) -> str:
        """Get the account ID for the Afero IoT account."""
        return self._account_id

    @property
    def afero_client(self) -> str:
        """Get identifier for Afero system."""
        return self._afero_client

    def add_controller(self, name: str, controller_type: type) -> None:
        """Add and instantiate a controller.

        The instantiated controller will be available as an attribute on the bridge
        instance with the provided name.

        :param name: The attribute name for the controller on the bridge instance.
        :param controller_type: The class of the controller to instantiate.
        """
        self._controllers[name] = controller_type(self)
        setattr(self, name, self._controllers[name])

    def set_token_data(self, data: TokenData) -> None:
        """Set TokenData used for querying the API.

        :param data: The `TokenData` object to set.
        """
        self._auth.set_token_data(data)

    def set_polling_interval(self, polling_interval: int) -> None:
        """Set the time between polling Afero API.

        :param polling_interval: The polling interval in seconds.
        """
        self._events.polling_interval = polling_interval

    def generate_api_url(self, endpoint: str) -> str:
        """Generate a URL for the Afero API.

        :param endpoint: The API endpoint path.

        :return: The fully qualified API URL.
        """
        endpoint = endpoint.removeprefix("/")
        return f"https://{v1_const.AFERO_CLIENTS[self._afero_client]['API_HOST']}/{endpoint}"

    async def close(self) -> None:
        """Close connection and clean up resources."""
        for task in self._scheduled_tasks:
            task.cancel()
            await task
        self._scheduled_tasks = []
        await self.events.stop()
        if self._close_session and self._web_session:
            await self._web_session.close()
        self.logger.info("Connection to bridge closed.")

    def subscribe(
        self,
        callback: EventCallBackType,
    ) -> Callable:
        """Subscribe to status changes for all resources.

        :param callback: The function to call when an event occurs.

        :return: A function that can be called to unsubscribe from the events.
        """
        unsubscribes = [
            controller.subscribe(callback) for controller in self.controllers
        ]

        def unsubscribe():
            for unsub in unsubscribes:
                unsub()

        return unsubscribe

    async def get_account_id(self) -> str:
        """Lookup the account ID associated with the login.

        :return: The account ID.
        :raises AferoError: If no account ID is found in the API response.
        """
        if not self._account_id:
            self.logger.debug("Querying API for account id")
            headers = {"host": v1_const.AFERO_CLIENTS[self._afero_client]["API_HOST"]}
            url = self.generate_api_url(v1_const.AFERO_GENERICS["ACCOUNT_ID_ENDPOINT"])
            with self.secret_logger():
                self.logger.debug(
                    "GETURL: %s, Headers: %s",
                    url,
                    headers,
                )
            res = await self.request(
                "GET",
                url,
                headers=headers,
            )
            res.raise_for_status()
            json_data = await res.json()
            if len(json_data) == 0 or len(json_data.get("accountAccess", [])) == 0:
                raise AferoError("No account ID found")
            self._account_id = (
                json_data.get("accountAccess")[0].get("account").get("accountId")
            )
            add_secret(self._account_id)
        return self._account_id

    async def initialize(self) -> None:
        """Initialize the bridge for communication with Afero API.

        To ensure the bridge is fully initialized, call async_block_until_done().
        """
        if len(self._scheduled_tasks) == 0:
            await self.get_account_id()
            for controller in self._controllers.values():
                if controller.initialized:
                    continue
                self.add_job(asyncio.create_task(controller.initialize()))
            self.add_job(asyncio.create_task(self.initialize_cleanup()))
            self.add_job(asyncio.create_task(self.events.initialize()))
            self.add_job(asyncio.create_task(self.events.wait_for_first_poll()))

    async def fetch_data(self, version_poll=False) -> list[dict[Any, str]]:
        """Query the API for all device data.

        :param version_poll: If True, also poll for device version information.

        :return: A list of dictionaries, each representing a device.
        """
        task = asyncio.create_task(self._fetch_data(version_poll))
        self.add_job(task)
        await task
        return task.result()

    async def _fetch_data(self, version_poll=False) -> list[dict[Any, str]]:
        """Query the API."""
        self.logger.debug("Querying API for all data")
        headers = {
            "host": v1_const.AFERO_CLIENTS[self._afero_client]["API_DATA_HOST"],
        }
        params = {"expansions": "state,capabilities,semantics"}
        if self.temperature_unit == TemperatureUnit.FAHRENHEIT:
            params["units"] = self.temperature_unit.value
        url = self.generate_api_url(
            v1_const.AFERO_GENERICS["API_DEVICE_ENDPOINT"].format(self.account_id)
        )
        res = await self.request(
            "get",
            url,
            headers=headers,
            params=params,
        )
        res.raise_for_status()
        data = await res.json()
        if not isinstance(data, list):
            raise TypeError(data)
        if version_poll:
            devs = {}
            for dev in data:
                if dev.get("typeId") != "metadevice.device":
                    continue
                dev_id = dev.get("deviceId")
                if dev_id in devs:
                    dev["version_data"] = devs[dev_id]
                    continue
                dev["version_data"] = await self.get_device_version(dev_id)
                devs[dev_id] = dev["version_data"]

        return data

    async def fetch_device_states(self, device_id) -> list[dict[Any, str]]:
        """Query the API for new device states.

        :param device_id: The ID of the device to fetch states for.

        :return: A list of `AferoState` objects representing the device's states.
        """
        task = asyncio.create_task(self._fetch_device_states(device_id))
        self.add_job(task)
        await task
        return task.result()

    async def _fetch_device_states(self, device_id) -> list[dict[Any, str]]:
        """Query the API for new device states."""
        self.logger.debug("Querying the API for updated states for %s", device_id)
        headers = {
            "host": v1_const.AFERO_CLIENTS[self._afero_client]["API_DATA_HOST"],
        }
        params = {"expansions": "state,capabilities,semantics"}
        url = self.generate_api_url(
            v1_const.AFERO_GENERICS["API_DEVICE_STATE_ENDPOINT"].format(
                self.account_id, device_id
            )
        )
        res = await self.request(
            "get",
            url,
            headers=headers,
            params=params,
        )
        res.raise_for_status()
        data = await res.json()
        states = []
        for state in data.get("values", []):
            try:
                states.append(AferoState(**state))
            except TypeError:
                continue
        return states

    async def get_device_version(self, device_id: str) -> dict:
        """Query the API for device version information.

        :param device_id: The ID of the device to get version info for.

        :return: A dictionary containing version information.
        """
        endpoint = v1_const.AFERO_GENERICS["API_DEVICE_VERSIONS_ENDPOINT"].format(
            self.account_id, device_id
        )
        url = self.generate_api_url(endpoint)
        res = await self.request("GET", url)
        res.raise_for_status()
        return await res.json()

    @asynccontextmanager
    async def create_request(
        self, method: str, url: str, include_token: bool, **kwargs
    ) -> Generator[aiohttp.ClientResponse, None, None]:
        """Create and manage an `aiohttp` request.

        This is an async context manager that handles session creation and
        authentication headers.

        :param method: The HTTP method (e.g., "GET", "POST").
        :param url: The URL for the request.
        :param include_token: If True, an Authorization header with a bearer token
            will be included.
        """
        if self._web_session is None:
            connector = aiohttp.TCPConnector(
                limit_per_host=3,
            )
            self._web_session = aiohttp.ClientSession(connector=connector)

        extras = {}
        if include_token:
            try:
                extras["Authorization"] = f"Bearer {await self._auth.token()}"
            except InvalidAuth:
                self.events.emit(EventType.INVALID_AUTH)
                raise
        headers = self.get_headers(**extras)
        headers.update(kwargs.get("headers", {}))
        kwargs["headers"] = headers
        kwargs["ssl"] = True
        async with self._web_session.request(method, url, **kwargs) as res:
            yield res

    async def request(
        self, method: str, url: str, include_token: bool = True, **kwargs
    ) -> aiohttp.ClientResponse:
        """Make a request to the API with automatic retries.

        :param method: The HTTP method.
        :param url: The request URL.
        :param include_token: Whether to include the auth token. Defaults to True.

        :return: The `aiohttp.ClientResponse` object.
        :raises ExceededMaximumRetries: If the request fails after all retries.
        """
        retries = 0
        with self.secret_logger():
            self.logger.info("Making request [%s] to %s with %s", method, url, kwargs)
        while retries < v1_const.MAX_RETRIES:
            retries += 1
            if retries > 1:
                retry_wait = 0.25 * retries
                await asyncio.sleep(retry_wait)
            async with self.create_request(
                method, url, include_token, **kwargs
            ) as resp:
                # 504 means the API is overloaded, back off a bit
                # 503 means the service is temporarily unavailable, back off a bit.
                # 429 means the bridge is rate limiting/overloaded, we should back off a bit.
                if resp.status in [429, 503, 504]:
                    continue
                # 403 is bad auth
                if resp.status == 403:
                    raise web_exceptions.HTTPForbidden
                await resp.read()
                return resp
        raise ExceededMaximumRetries("Exceeded maximum number of retries")

    async def send_service_request(self, device_id: str, states: list[dict[str, Any]]):
        """Manually send state requests to Afero IoT.

        :param device_id: ID for the device
        :param states: List of states to send

        :raises DeviceNotFound: If the device with the given ID is not found.
        """
        controller = self._known_devs.get(device_id)
        if not controller:
            raise DeviceNotFound(f"Unable to find device {device_id}")
        await controller.update(device_id, states=states)

    def get_headers(self, **kwargs):
        """Get default headers for an API call.

        :param kwargs: Additional headers to include.
        :return: A dictionary of headers.
        """
        headers: dict[str, str] = {
            "user-agent": v1_const.AFERO_GENERICS["DEFAULT_USERAGENT"].safe_substitute(
                client_name=self.client_name
            ),
            "accept-encoding": "gzip",
        }
        headers.update(kwargs)
        return headers

    # Task management enables us to block until finished
    def add_job(self, task: asyncio.Task) -> None:
        """Add a job to be processed."""
        self._adhoc_tasks.append(task)

    async def async_block_until_done(self):
        """Block until all ad-hoc and event queue tasks are completed."""
        await asyncio.gather(*self._adhoc_tasks)
        await self.events.async_block_until_done()

    async def initialize_cleanup(self) -> None:
        """Start the background task that cleans up completed ad-hoc tasks."""
        self._scheduled_tasks.append(asyncio.create_task(self.__cleanup_processor()))

    async def __cleanup_processor(self) -> None:
        """Periodically remove finished tasks from the ad-hoc task list."""
        with contextlib.suppress(asyncio.CancelledError):
            while True:
                for task in self._adhoc_tasks[:]:
                    if task.done():
                        self._adhoc_tasks.remove(task)
                await asyncio.sleep(1)

    async def adjust_temperature_unit(
        self,
        temperature_unit: TemperatureUnit,
    ) -> None:
        """Adjust the temperature unit for API responses.

        :param temperature_unit: The desired temperature unit for API responses.
        """
        if self.temperature_unit != temperature_unit:
            self.temperature_unit = temperature_unit
            self.add_job(asyncio.create_task(self.events.perform_poll()))



from dataclasses import replace

import pytest

from aioafero import anonomyize_data
from aioafero.device import AferoDevice, AferoState, AferoCapability

POWER_STATE = {
    "functionClass": "power",
    "functionInstance": None,
    "value": "on",
    "lastUpdateTime": 0,
}
GEO_STATE = {
    "functionClass": "geo-coordinates",
    "functionInstance": None,
    "value": {"geo-coordinates": {"latitude": "5", "longitude": "5"}},
}
MAC_STATE = {
    "functionClass": "wifi-mac-address",
    "functionInstance": None,
    "value": "aa:bb:cc:dd:ee:ff",
}
child_dev_1 = AferoDevice(
    id="c-id-1",
    device_id="parent-1",
    model="test-c-dev-1",
    device_class="test-c-dev-1",
    default_name="test-c-dev-1",
    default_image="test-c-dev-1",
    friendly_name="test-c-dev-1",
    functions=[],
    states=[
        AferoState(**POWER_STATE),
        AferoState(**GEO_STATE),
        AferoState(**MAC_STATE),
    ],
    children=[],
    manufacturerName="test-manuf",
)
child_dev_2 = AferoDevice(
    id="c-id-2",
    device_id="parent-1",
    model="test-c-dev-2",
    device_class="test-c-dev-2",
    default_name="test-c-dev-2",
    default_image="test-c-dev-2",
    friendly_name="test-c-dev-2",
    functions=[],
    states=[
        AferoState(**POWER_STATE),
        AferoState(**GEO_STATE),
        AferoState(**MAC_STATE),
    ],
    children=[],
    manufacturerName="test-manuf",
)

parent_dev_1 = AferoDevice(
    id="p-id-1",
    device_id="parent-1",
    model="test-dev-1",
    device_class="test-dev-1",
    default_name="test-dev-1",
    default_image="test-dev-1",
    friendly_name="test-dev-1",
    capabilities=[
        AferoCapability(
            functionClass="sensor-state",
            type="object",
            schedulable=False,
            functionInstance="sensor-1",
            _opts={
                "name": "Aaaaa",
                "locale": "en_US"
            }
        )
    ],
    functions=[],
    states=[
        AferoState(**POWER_STATE),
        AferoState(**GEO_STATE),
        AferoState(**MAC_STATE),
    ],
    children=[child_dev_1.id, child_dev_2.id],
    manufacturerName="test-manuf",
)


@pytest.fixture
def mock_uuid(mocker):
    calls = 0

    def get_uuid():
        nonlocal calls
        calls += 1
        if calls % 2:
            return "its-a-1"
        return "its-a-2"

    mocker.patch.object(anonomyize_data, "uuid4", side_effect=get_uuid)


@pytest.mark.parametrize(
    "state,only_geo,expected",
    [
        # Non-geo + only_geo
        (
            AferoState(
                functionClass="wifi-mac-address",
                functionInstance=None,
                value="aa:bb:cc:dd:ee:ff",
            ),
            True,
            {
                "functionClass": "wifi-mac-address",
                "functionInstance": None,
                "lastUpdateTime": 0,
                "value": "aa:bb:cc:dd:ee:ff",
            },
        ),
        # Non-geo + not only_geo
        (
            AferoState(
                functionClass="wifi-mac-address",
                functionInstance=None,
                value="aa:bb:cc:dd:ee:ff",
            ),
            False,
            {
                "functionClass": "wifi-mac-address",
                "functionInstance": None,
                "lastUpdateTime": 0,
                "value": "anon data",
            },
        ),
        # Geo data
        (
            AferoState(
                functionClass="geo-coordinates",
                functionInstance=None,
                value={"geo-coordinates": {"latitude": "5", "longitude": "5"}},
            ),
            False,
            {
                "functionClass": "geo-coordinates",
                "functionInstance": None,
                "lastUpdateTime": 0,
                "value": {"geo-coordinates": {"latitude": "0", "longitude": "0"}},
            },
        ),
        # Non-anon data
        (
            AferoState(
                functionClass="power",
                functionInstance=None,
                lastUpdateTime=12345,
                value="on",
            ),
            False,
            {
                "functionClass": "power",
                "functionInstance": None,
                "lastUpdateTime": 0,
                "value": "on",
            },
        ),
    ],
)
def test_anonymize_state(state, only_geo, expected, mocker):
    mocker.patch.object(anonomyize_data, "uuid4", return_value="anon data")
    assert anonomyize_data.anonymize_state(state, only_geo=only_geo) == expected


@pytest.mark.parametrize(
    "device, device_links, parent_mapping, anon_name, expected",
    [
        # Anon name
        (
            replace(child_dev_1),
            {},
            {},
            True,
            {
                "split_identifier": None,
                "friendly_name": "friendly-device-0",
                "id": "its-a-1",
                "device_id": "its-a-2",
                "states": [
                    POWER_STATE,
                    {
                        "functionClass": "geo-coordinates",
                        "functionInstance": None,
                        "lastUpdateTime": 0,
                        "value": {
                            "geo-coordinates": {"latitude": "0", "longitude": "0"}
                        },
                    },
                    {
                        "functionClass": "wifi-mac-address",
                        "functionInstance": None,
                        "lastUpdateTime": 0,
                        "value": "its-a-1",
                    },
                ],
            },
        ),
        # normal name
        (
            replace(child_dev_1),
            {child_dev_1.device_id: "anon-device-id"},
            {child_dev_1.id: {"parent": parent_dev_1.id, "new": "anon-id"}},
            False,
            {
                "split_identifier": None,
                "friendly_name": child_dev_1.friendly_name,
                "id": "anon-id",
                "device_id": "anon-device-id",
                "states": [
                    POWER_STATE,
                    {
                        "functionClass": "geo-coordinates",
                        "functionInstance": None,
                        "lastUpdateTime": 0,
                        "value": {
                            "geo-coordinates": {"latitude": "0", "longitude": "0"}
                        },
                    },
                    {
                        "functionClass": "wifi-mac-address",
                        "functionInstance": None,
                        "lastUpdateTime": 0,
                        "value": "its-a-1",
                    },
                ],
            },
        ),
    ],
)
def test_anonymize_device(
    device, device_links, parent_mapping, anon_name, expected, mock_uuid
):
    anon_dev = anonomyize_data.anonymize_device(
        device, parent_mapping, device_links, anon_name
    )
    assert device.device_id in device_links
    for key, val in expected.items():
        assert key in anon_dev
        assert anon_dev[key] == val


@pytest.mark.parametrize(
    "devices, expected, new_children",
    [
        # No parents
        ([child_dev_1], {}, []),
        # Parents
        (
            [parent_dev_1, child_dev_1, child_dev_2],
            {
                "c-id-1": {"parent": "its-a-1", "new": "its-a-2"},
                "c-id-2": {"parent": "its-a-1", "new": "its-a-1"},
            },
            ["its-a-2", "its-a-1"],
        ),
    ],
)
def test_generate_parent_mapping(devices, expected, new_children, mock_uuid):
    assert anonomyize_data.generate_parent_mapping(devices) == expected
    if new_children:
        assert devices[0].children == new_children


@pytest.mark.parametrize("anon_name", [True, False])
def test_anonymize_devices(anon_name, mock_uuid):
    expected = [
        {
            "split_identifier": None,
            "id": "its-a-1",
            "device_id": "its-a-2",
            "model": "test-dev-1",
            "device_class": "test-dev-1",
            "default_name": "test-dev-1",
            "default_image": "test-dev-1",
            "friendly_name": "test-dev-1",
            "capabilities": [
                {
                '_opts': {
                    'locale': 'en_US',
                    'name': 'Aaaaa',
                },
                'functionClass': 'sensor-state',
                'functionInstance': 'sensor-1',
                'schedulable': False,
                'type': 'object',
                },
            ],
            "functions": [],
            "states": [
                {
                    "functionClass": "power",
                    "value": "on",
                    "lastUpdateTime": 0,
                    "functionInstance": None,
                },
                {
                    "functionClass": "geo-coordinates",
                    "value": {"geo-coordinates": {"latitude": "0", "longitude": "0"}},
                    "lastUpdateTime": 0,
                    "functionInstance": None,
                },
                {
                    "functionClass": "wifi-mac-address",
                    "value": "its-a-1",
                    "lastUpdateTime": 0,
                    "functionInstance": None,
                },
            ],
            "children": ["its-a-2", "its-a-1"],
            "manufacturerName": "test-manuf",
            'version_data': None,
        },
        {
            "split_identifier": None,
            "id": "its-a-2",
            "device_id": "its-a-2",
            "model": "test-c-dev-1",
            "device_class": "test-c-dev-1",
            "default_name": "test-c-dev-1",
            "default_image": "test-c-dev-1",
            "friendly_name": "test-c-dev-1",
            "capabilities": [],
            "functions": [],
            "states": [
                {
                    "functionClass": "power",
                    "value": "on",
                    "lastUpdateTime": 0,
                    "functionInstance": None,
                },
                {
                    "functionClass": "geo-coordinates",
                    "value": {"geo-coordinates": {"latitude": "0", "longitude": "0"}},
                    "lastUpdateTime": 0,
                    "functionInstance": None,
                },
                {
                    "functionClass": "wifi-mac-address",
                    "value": "its-a-1",
                    "lastUpdateTime": 0,
                    "functionInstance": None,
                },
            ],
            "children": [],
            "manufacturerName": "test-manuf",
            'version_data': None,
        },
        {
            "split_identifier": None,
            "id": "its-a-2",
            "device_id": "its-a-2",
            "model": "test-c-dev-2",
            "device_class": "test-c-dev-2",
            "default_name": "test-c-dev-2",
            "default_image": "test-c-dev-2",
            "friendly_name": "test-c-dev-2",
            "capabilities": [],
            "functions": [],
            "states": [
                {
                    "functionClass": "power",
                    "value": "on",
                    "lastUpdateTime": 0,
                    "functionInstance": None,
                },
                {
                    "functionClass": "geo-coordinates",
                    "value": {"geo-coordinates": {"latitude": "0", "longitude": "0"}},
                    "lastUpdateTime": 0,
                    "functionInstance": None,
                },
                {
                    "functionClass": "wifi-mac-address",
                    "value": "its-a-1",
                    "lastUpdateTime": 0,
                    "functionInstance": None,
                },
            ],
            "children": [],
            "manufacturerName": "test-manuf",
            'version_data': None,
        },
    ]
    if anon_name:
        expected[0]["friendly_name"] = "friendly-device-1"
        expected[1]["friendly_name"] = "friendly-device-2"
        expected[2]["friendly_name"] = "friendly-device-3"
    assert (
        anonomyize_data.anonymize_devices(
            [parent_dev_1, child_dev_1, child_dev_2], anon_name=anon_name
        )
        == expected
    )



"""Controller holding and managing Hubspace resources of type `light`."""

from contextlib import suppress
import copy

from aioafero import device, errors
from aioafero.device import AferoDevice, AferoState
from aioafero.util import process_range
from aioafero.v1.models import features
from aioafero.v1.models.light import Light, LightPut
from aioafero.v1.models.resource import DeviceInformation, ResourceTypes

from .base import AferoBinarySensor, AferoSensor, BaseResourcesController
from .event import CallbackResponse

SPLIT_IDENTIFIER: str = "light"


def process_names(values: list[dict]) -> set[str]:
    """Extract unique names from the elements."""
    vals = set()
    for val in values:
        vals.add(val["name"])
    return vals


def generate_split_name(afero_device: AferoDevice, instance: str) -> str:
    """Generate the name for an instanced element."""
    return f"{afero_device.id}-{SPLIT_IDENTIFIER}-{instance}"


def get_split_instances(afero_dev: AferoDevice) -> list[tuple[str, ResourceTypes]]:
    """Determine available instances from the states."""
    instances = set()
    if afero_dev.model == "LCN3002LM-01 WH":
        instances.add(("color", ResourceTypes.LIGHT))
        instances.add(("white", ResourceTypes.LIGHT))
    for state in afero_dev.states:
        if state.functionClass == "toggle" and state.functionInstance not in [
            x[0] for x in instances
        ]:
            instances.add((state.functionInstance, ResourceTypes.SWITCH))
    return sorted(instances)


def get_valid_states(afero_dev: AferoDevice, instance: str) -> list:
    """Find states associated with the specific instance."""
    valid_states: list = []
    if not any(state.functionClass == "toggle" for state in afero_dev.states):
        return valid_states
    for state in afero_dev.states:
        if state.functionClass == "available":
            valid_states.append(state)
        elif afero_dev.model == "LCN3002LM-01 WH":
            if state.functionInstance == "primary":
                continue
            if (instance == "white" and state.functionInstance == instance) or (
                instance != "white" and state.functionInstance != "white"
            ):
                valid_states.append(state)
        elif state.functionClass == "toggle":
            valid_states.append(state)
    return valid_states


def light_callback(afero_device: AferoDevice) -> CallbackResponse:
    """Convert an AferoDevice into multiple devices."""
    multi_devs: list[AferoDevice] = []
    if afero_device.device_class == ResourceTypes.LIGHT.value:
        children = []
        for instance, resource_type in get_split_instances(afero_device):
            cloned = copy.deepcopy(afero_device)
            cloned.device_class = resource_type.value
            cloned.id = generate_split_name(afero_device, instance)
            cloned.split_identifier = SPLIT_IDENTIFIER
            cloned.friendly_name = f"{afero_device.friendly_name} - {instance}"
            cloned.states = get_valid_states(afero_device, instance)
            cloned.children = []
            multi_devs.append(cloned)
            children.append(cloned.id)
        if afero_device.model == "LCN3002LM-01 WH":
            cloned = copy.deepcopy(afero_device)
            valid_states = [
                state
                for state in afero_device.states
                if state.functionClass
                in [
                    "available",
                    "wifi-ssid",
                    "wifi-rssi",
                    "wifi-steady-state",
                    "wifi-setup-state",
                    "wifi-mac-address",
                    "ble-mac-address",
                ]
            ]
            cloned.states = valid_states
            cloned.children = children
            cloned.device_class = "parent-device"
            multi_devs.append(cloned)
    return CallbackResponse(
        split_devices=multi_devs,
        remove_original=(afero_device.model == "LCN3002LM-01 WH"),
    )


class LightController(BaseResourcesController[Light]):
    """Controller holding and managing Hubspace resources of type `light`."""

    ITEM_TYPE_ID = ResourceTypes.DEVICE
    ITEM_TYPES = [ResourceTypes.LIGHT]
    ITEM_CLS = Light
    ITEM_MAPPING = {
        "color": "color-rgb",
        "color_mode": "color-mode",
        "color_temperature": "color-temperature",
        "dimming": "brightness",
        "effect": "color-sequence",
    }
    ITEM_SELECTS = {("speed", "color-sequence"): "Speed"}
    # Split Lights from the primary payload
    DEVICE_SPLIT_CALLBACKS: dict[str, callable] = {
        ResourceTypes.LIGHT.value: light_callback
    }

    async def turn_on(self, device_id: str) -> None:
        """Turn on the light."""
        await self.set_state(device_id, on=True)

    async def turn_off(self, device_id: str) -> None:
        """Turn off the light."""
        await self.set_state(device_id, on=False)

    async def set_color_temperature(self, device_id: str, temperature: int) -> None:
        """Set Color Temperature to light. Turn on light if it's currently off."""
        await self.set_state(
            device_id, on=True, temperature=temperature, color_mode="white"
        )

    async def set_brightness(self, device_id: str, brightness: int) -> None:
        """Set brightness of the light. Turn on light if it's currently off."""
        await self.set_state(device_id, on=True, brightness=brightness)

    async def set_rgb(self, device_id: str, red: int, green: int, blue: int) -> None:
        """Set RGB of the light. Turn on light if it's currently off."""
        await self.set_state(
            device_id, on=True, color=(red, green, blue), color_mode="color"
        )

    async def set_effect(self, device_id: str, effect: str) -> None:
        """Set effect of the light. Turn on light if it's currently off."""
        await self.set_state(device_id, on=True, effect=effect, color_mode="sequence")

    async def initialize_elem(self, afero_device: AferoDevice) -> Light:
        """Initialize the element.

        :param afero_device: Afero Device that contains the updated states

        :return: Newly initialized resource
        """
        available: bool = False
        on: features.OnFeature | None = None
        color_temp: features.ColorTemperatureFeature | None = None
        color: features.ColorFeature | None = None
        color_mode: features.ColorModeFeature | None = None
        dimming: features.DimmingFeature | None = None
        effect: features.EffectFeature | None = None
        sensors: dict[str, AferoSensor] = {}
        binary_sensors: dict[str, AferoBinarySensor] = {}
        selects: dict[tuple[str, str], features.SelectFeature] = {}
        for state in afero_device.states:
            func_def = device.get_function_from_device(
                afero_device.functions, state.functionClass, state.functionInstance
            )
            if state.functionClass == "power" or (
                afero_device.split_identifier and state.functionClass == "toggle"
            ):
                on = features.OnFeature(
                    on=state.value == "on",
                    func_class=state.functionClass,
                    func_instance=state.functionInstance,
                )
            elif state.functionClass == "color-temperature":
                if len(func_def["values"]) > 1:
                    avail_temps = process_color_temps(func_def["values"])
                else:
                    avail_temps = process_range(func_def["values"][0])
                prefix = "K" if func_def.get("type", None) != "numeric" else ""
                current_temp = state.value
                if isinstance(current_temp, str) and current_temp.endswith("K"):
                    current_temp = current_temp[:-1]
                color_temp = features.ColorTemperatureFeature(
                    temperature=int(current_temp), supported=avail_temps, prefix=prefix
                )
            elif state.functionClass == "brightness":
                temp_bright = process_range(func_def["values"][0])
                dimming = features.DimmingFeature(
                    brightness=int(state.value), supported=temp_bright
                )
            elif state.functionClass == "color-sequence":
                current_effect = state.value
                effects = process_effects(afero_device.functions)
                effect = features.EffectFeature(effect=current_effect, effects=effects)
            elif state.functionClass == "color-rgb":
                color = features.ColorFeature(
                    red=state.value["color-rgb"].get("r", 0),
                    green=state.value["color-rgb"].get("g", 0),
                    blue=state.value["color-rgb"].get("b", 0),
                )
            elif state.functionClass == "color-mode":
                color_mode = features.ColorModeFeature(state.value)
            elif state.functionClass == "available":
                available = state.value
            elif select := await self.initialize_select(afero_device.functions, state):
                selects[select[0]] = select[1]

        supported_color_modes: list[str] = []
        for function in afero_device.functions:
            if function["functionClass"] != "color-mode":
                continue
            supported_color_modes = [
                supported_color_mode["name"]
                for supported_color_mode in function["values"]
            ]
            break

        self._items[afero_device.id] = Light(
            _id=afero_device.id,
            available=available,
            split_identifier=afero_device.split_identifier,
            sensors=sensors,
            binary_sensors=binary_sensors,
            device_information=DeviceInformation(
                device_class=afero_device.device_class,
                default_image=afero_device.default_image,
                default_name=afero_device.default_name,
                manufacturer=afero_device.manufacturerName,
                model=afero_device.model,
                name=afero_device.friendly_name,
                parent_id=afero_device.device_id,
                children=afero_device.children,
                functions=afero_device.functions,
            ),
            on=on,
            dimming=dimming,
            color_mode=color_mode,
            color_temperature=color_temp,
            color=color,
            color_modes=supported_color_modes,
            effect=effect,
            selects=selects,
        )
        return self._items[afero_device.id]

    async def update_elem(self, afero_device: AferoDevice) -> set:
        """Update the Light with the latest API data.

        :param afero_device: Afero Device that contains the updated states

        :return: States that have been modified
        """
        cur_item = self.get_device(afero_device.id)
        updated_keys = set()
        color_seq_states: dict[str, AferoState] = {}
        for state in afero_device.states:
            if state.functionClass == "power" or (
                afero_device.split_identifier and state.functionClass == "toggle"
            ):
                new_val = state.value == "on"
                if cur_item.on.on != new_val:
                    cur_item.on.on = new_val
                    updated_keys.add("on")
            elif state.functionClass == "color-temperature":
                current_temp = state.value
                if isinstance(current_temp, str) and current_temp.endswith("K"):
                    current_temp = current_temp[:-1]
                new_val = int(current_temp)
                if cur_item.color_temperature.temperature != new_val:
                    cur_item.color_temperature.temperature = new_val
                    updated_keys.add("color_temperature")
            elif state.functionClass == "brightness":
                new_val = int(state.value)
                if cur_item.dimming.brightness != new_val:
                    cur_item.dimming.brightness = int(state.value)
                    updated_keys.add("dimming")
            elif state.functionClass == "color-sequence":
                color_seq_states[state.functionInstance] = state
            elif state.functionClass == "color-rgb":
                color_red = state.value["color-rgb"].get("r", 0)
                color_green = state.value["color-rgb"].get("g", 0)
                color_blue = state.value["color-rgb"].get("b", 0)
                if (
                    cur_item.color.red != color_red
                    or cur_item.color.green != color_green
                    or cur_item.color.blue != color_blue
                ):
                    cur_item.color.red = color_red
                    cur_item.color.green = color_green
                    cur_item.color.blue = color_blue
                    updated_keys.add("color")
            elif state.functionClass == "color-mode":
                if cur_item.color_mode.mode != state.value:
                    cur_item.color_mode.mode = state.value
                    updated_keys.add("color_mode")
            elif state.functionClass == "available":
                if cur_item.available != state.value:
                    cur_item.available = state.value
                    updated_keys.add("available")
            elif update_key := await self.update_select(state, cur_item):
                updated_keys.add(update_key)

        # Several states hold the effect, but its always derived from the preset functionInstance
        return updated_keys.union(
            await self.update_elem_color(cur_item, color_seq_states)
        )

    async def update_elem_color(self, cur_item: Light, color_seq_states: dict) -> set:
        """Perform the update for effects."""
        updated_keys = set()
        if color_seq_states and "preset" in color_seq_states:
            preset_val = color_seq_states["preset"].value
            if cur_item.effect.is_preset(preset_val):
                if cur_item.effect.effect != preset_val:
                    cur_item.effect.effect = preset_val
                    updated_keys.add("effect")
            else:
                new_val = color_seq_states[color_seq_states["preset"].value].value
                if cur_item.effect.effect != new_val:
                    cur_item.effect.effect = color_seq_states[
                        color_seq_states["preset"].value
                    ].value
                    updated_keys.add("effect")
        return updated_keys

    async def set_state(
        self,
        device_id: str,
        on: bool | None = None,
        temperature: int | None = None,
        brightness: int | None = None,
        color_mode: str | None = None,
        color: tuple[int, int, int] | None = None,
        effect: str | None = None,
        force_white_mode: int | None = None,
        selects: dict[tuple[str, str | None], int | str] | None = None,
    ) -> None:
        """Set supported feature(s) to light resource.

        force_white_mode's value should be the brightness percentage after switching to white
        """
        update_obj = LightPut()
        try:
            cur_item = self.get_device(device_id)
        except errors.DeviceNotFound:
            self._logger.info("Unable to find device %s", device_id)
            return
        if on is not None:
            update_obj.on = features.OnFeature(
                on=on,
                func_class=cur_item.on.func_class,
                func_instance=cur_item.on.func_instance,
            )
        send_duplicate_states = False
        if force_white_mode is not None:
            send_duplicate_states = True
            update_obj.color_mode = features.ColorModeFeature(mode="white")
            update_obj.dimming = features.DimmingFeature(
                brightness=force_white_mode, supported=cur_item.dimming.supported
            )
        else:
            if selects:
                for key, val in selects.items():
                    if key not in cur_item.selects:
                        continue
                    update_obj.selects[key] = features.SelectFeature(
                        selected=val,
                        selects=cur_item.selects[key].selects,
                        name=cur_item.selects[key].name,
                    )
            if temperature is not None and cur_item.color_temperature is not None:
                adjusted_temp = min(
                    cur_item.color_temperature.supported,
                    key=lambda x: abs(x - temperature),
                )
                update_obj.color_temperature = features.ColorTemperatureFeature(
                    temperature=adjusted_temp,
                    supported=cur_item.color_temperature.supported,
                    prefix=cur_item.color_temperature.prefix,
                )
            if brightness is not None and cur_item.dimming is not None:
                update_obj.dimming = features.DimmingFeature(
                    brightness=brightness, supported=cur_item.dimming.supported
                )
            if color is not None and cur_item.color is not None:
                update_obj.color = features.ColorFeature(
                    red=color[0], green=color[1], blue=color[2]
                )
            if color_mode is not None and cur_item.color_mode is not None:
                update_obj.color_mode = features.ColorModeFeature(mode=color_mode)
            if effect is not None and cur_item.effect is not None:
                update_obj.effect = features.EffectFeature(
                    effect=effect, effects=cur_item.effect.effects
                )
        await self.update(
            device_id, obj_in=update_obj, send_duplicate_states=send_duplicate_states
        )


def process_color_temps(color_temps: dict) -> list[int]:
    """Determine the supported color temps.

    :param color_temps: Result from functions["values"]
    """
    supported_temps = []
    for temp in color_temps:
        color_temp = temp["name"]
        if isinstance(color_temp, str) and color_temp.endswith("K"):
            color_temp = color_temp[0:-1]
        supported_temps.append(int(color_temp))
    return sorted(supported_temps)


def process_effects(functions: list[dict]) -> dict[str, set]:
    """Determine the supported effects."""
    supported_effects = {}
    for function in functions:
        if function["functionClass"] == "color-sequence":
            supported_effects[function["functionInstance"]] = set(
                process_names(function["values"])
            )
    # custom shouldn't be a value in preset
    with suppress(KeyError):
        supported_effects["preset"].remove("custom")
    return supported_effects



"""Generic utils for interacting with Afero IoT API or the responses."""

from typing import Any


def percentage_to_ordered_list_item[_T](ordered_list: list[_T], percentage: int) -> _T:
    """Find the item that most closely matches the percentage in an ordered list.

    When using this utility for fan speeds, do not include "off"

    Given the list: ["low", "medium", "high", "very_high"], this
    function will return the following when the item is passed
    in:

        1-25: low
        26-50: medium
        51-75: high
        76-100: very_high
    """
    if not (list_len := len(ordered_list)):
        raise ValueError("The ordered list is empty")

    for offset, speed in enumerate(ordered_list):
        list_position = offset + 1
        upper_bound = (list_position * 100) // list_len
        if percentage <= upper_bound:
            return speed

    return ordered_list[-1]


def ordered_list_item_to_percentage[_T](ordered_list: list[_T], item: _T) -> int:
    """Determine the percentage of an item in an ordered list.

    When using this utility for fan speeds, do not include "off"

    Given the list: ["low", "medium", "high", "very_high"], this
    function will return the following when the item is passed
    in:

        low: 25
        medium: 50
        high: 75
        very_high: 100

    """
    if item not in ordered_list:
        raise ValueError(f'The item "{item}" is not in "{ordered_list}"')

    list_len = len(ordered_list)
    list_position = ordered_list.index(item) + 1
    return (list_position * 100) // list_len


def process_range(range_vals: dict) -> list[Any]:
    """Process a range to determine what's supported.

    :param range_vals: Result from functions["values"][x]
    """
    supported_range = []
    range_min = range_vals["range"]["min"]
    range_max = range_vals["range"]["max"]
    range_step = range_vals["range"]["step"]
    if range_min == range_max:
        supported_range.append(range_max)
    else:
        supported_range = list(float_range(range_min, range_max, range_step))
    if range_max not in supported_range:
        supported_range.append(range_max)
    return supported_range


def float_range(start, stop, step):
    """Create a generator that yields a range as float."""
    while start < stop:
        yield start
        start += step


def process_function(
    functions: list[dict], func_class: str, func_instance: str | None = None
) -> list[Any]:
    """Generate a list of whatever you are searching for.

    :param functions: List of functions for the given device
    :param func_class: functionClass to search
    :param func_instance: functionInstance to search
    """
    results = []
    for function in functions:
        if function["functionClass"] != func_class:
            continue
        if func_instance and function.get("functionInstance") != func_instance:
            continue
        if function["type"] == "numeric":
            results = process_range(function["values"][0])
        elif function["type"] == "category":
            for value in function["values"]:
                results.append(value["name"])
        break
    return results
